from typing import Callable, Dict, Union  # @UnusedImport

from thonny import get_workbench
from thonny.misc_utils import running_on_linux, running_on_windows
from thonny.workbench import BasicUiThemeSettings, CompoundUiThemeSettings


def scale(value) -> float:
    # dimensions in this module were designed with a 1.67 scale
    return get_workbench().scale(value / 1.67)

def _treeview_settings() -> BasicUiThemeSettings:
    light_blue = "#ADD8E6" 
    light_grey = "#D3D3D3"
    
    if running_on_linux():
        bg_sel_focus = light_blue
        fg = "black"
    else:
        bg_sel_focus = 'SystemHighlight'
        fg = 'SystemHighlightText'
    
    return {
        "Treeview" : {
            "configure" : {
                "font" : "TreeviewFont"
            },
            "map" : {
                "background" : [('selected', 'focus', bg_sel_focus),
                                ('selected', '!focus', light_grey)],
                "foreground" : [('selected', fg)]
            },
            "layout" : [
                # get rid of borders
                ('Treeview.treearea', {'sticky': 'nswe'})
            ]
        },
        "treearea" : {
            "configure" : {
                "borderwidth" : 0
            },
        }
    }

def _menubutton_settings() -> BasicUiThemeSettings:
    return {
        "TMenubutton" : {
            "configure" : {
                "padding" : scale(14)
            },
            "layout" : [
                ('Menubutton.dropdown', {'side': 'right', 'sticky': 'ns'}),
                ('Menubutton.button', {'children': [
                    #('Menubutton.padding', {'children': [
                        ('Menubutton.label', {'sticky': ''})
                    #], 'expand': '1', 'sticky': 'we'})
                ], 'expand': '1', 'sticky': 'nswe'})
            ]
        }
    }

def _paned_window_settings() -> BasicUiThemeSettings:
    return {
        "Sash" : {
            "configure" : {
                "sashthickness" : scale(10)
            }
        }
    }


def _menu_settings() -> BasicUiThemeSettings:
    return {
        "Menubar" : {
            "configure" : {
                "activeborderwidth" : 0,
                "relief" : "flat",
            }
        },
    }


def _text_settings() -> BasicUiThemeSettings:
    return {
        "Text" : {
            "configure" : {
                "background" : "SystemWindow" if running_on_windows() else "white",
                "foreground" : "SystemWindowText" if running_on_windows() else "black",
            },
        },
        "Syntax.Text" : {
            "map" : {
                "background" : [("readonly", "Yellow")]
            }
        },
        "Gutter" : {
            "configure" : {
                "background" : '#e0e0e0',
                "foreground" : '#999999' 
            }
        },
    }


def _label_settings() -> BasicUiThemeSettings:
    return {
        "Url.TLabel" : {
            "configure" : {
                "foreground" : "#3A66DD",
            },
        },
    }


def _button_notebook_settings() -> BasicUiThemeSettings:
    # Adapted from https://github.com/python/cpython/blob/2.7/Demo/tkinter/ttk/notebook_closebtn.py
    return {
        "closebutton" : {
            "element create" : (
                "image", "img_close",
                ("active", "pressed", "!disabled", "img_close_active"),
                ("active", "!disabled", "img_close_active"), 
                {"padding" : scale(2), "sticky" : ''}                
            )
        },
        "ButtonNotebook.TNotebook.Tab" : {
            "layout" : [
                ("Notebook.tab", {"sticky": "nswe", "children": [
                    ("Notebook.padding", {"side": "top", "sticky": "nswe", "children": [
                        ("Notebook.focus", {"side": "left", "sticky": "nswe", "children": [
                            ("Notebook.label", {"side": "left", "sticky": ''}),
                        ]}),
                        ("Notebook.closebutton", {"side": "right", "sticky": ''})
                    ]})
                ]})
            ]
        },
    }

def clam() -> BasicUiThemeSettings:
    # Transcribed from https://github.com/tcltk/tk/blob/master/library/ttk/clamTheme.tcl
    defaultfg =      "#000000"
    disabledfg =     "#999999"
    frame =          "#dcdad5"
    window =         "#ffffff"
    dark =           "#cfcdc8"
    darker =         "#bab5ab"
    darkest =        "#9e9a91"
    lighter =        "#eeebe7"
    selectbg =       "#4a6984"
    selectfg =       "#ffffff"
    altindicator =   "#5895bc"
    disabledaltindicator =    "#a0a0a0"
    
    return {
        "." : {
            "configure" : {
                "background"        : frame,
                "foreground"        : defaultfg,
                "bordercolor"       : darkest,
                "darkcolor"         : dark,
                "lightcolor"        : lighter,
                "troughcolor"       : darker,
                "selectbackground"  : selectbg,
                "selectforeground"  : selectfg,
                "selectborderwidth" : 0,
                "font" : "TkDefaultFont",
            },
            
            "map" : {
                "background" : [("disabled", frame), 
                                ("active", lighter)],
                "foreground" : [("disabled", disabledfg)],
                "selectbackground" : [("!focus", darkest)],
                "selectforeground" : [("!focus", "white")]
            },
        },
            
        "TButton" : {
            "configure" : {
                "anchor" : "center", 
                "width" : scale(11), 
                "padding" : scale(5), 
                "relief" : "raised"
            },
            "map" : {
                "background" : [("disabled", frame),
                                ("pressed", darker),
                                ("active", lighter)],
                "lightcolor" : [("pressed", darker)],
                "darkcolor"  : [("pressed", darker)],
                "bordercolor": [("alternate", "#000000")],
            }
        },
        
        "Toolbutton" : {
            "configure" : {
                "anchor" : "center",
                "padding" : scale(2),
                "relief" : "flat"
            },
            "map" : {
                "relief" : [("disabled",  "flat"),
                            ("selected", "sunken"),
                            ("pressed", "sunken"),
                            ("active", "raised")],
                "background" : [("disabled", frame),
                                ("pressed", darker),
                                ("active", lighter)],
                "lightcolor" : [("pressed", darker)],
                "darkcolor" : [("pressed", darker)]
            }
        },
        
        "TCheckbutton" : {
            "configure" : {
                "indicatorbackground" : "#ffffff",
                "indicatormargin" : [scale(1), scale(1), scale(6), scale(1)],
                "padding" :  scale(2),
            },
            "map" : {
                "indicatorbackground" : [("pressed", frame),
                                         ("!disabled", "alternate", altindicator),
                                         ("disabled", "alternate", disabledaltindicator),
                                         ("disabled", frame)]
            }
        },
        
        # TRadiobutton has same style as TCheckbutton
        "TRadiobutton" : {
            "configure" : {
                "indicatorbackground" : "#ffffff",
                "indicatormargin" : [scale(1), scale(1), scale(6), scale(1)],
                "padding" :  scale(2),
            },
            "map" : {
                "indicatorbackground" : [("pressed", frame),
                                         ("!disabled", "alternate", altindicator),
                                         ("disabled", "alternate", disabledaltindicator),
                                         ("disabled", frame)]
            }
        },
        
        "TMenubutton" : {
            "configure" : {
                "width" : scale(11),
                "padding" : scale(5),
                "relief" : "raised"
            }
        },
        
        "TEntry" : {
            "configure" : {
                "padding" : scale(1),
                "insertwidth" : scale(1)
            },
            "map" : {
                "background" : [("readonly", frame)],
                "bordercolor" : [("focus", selectbg)],
                "lightcolor" : [("focus", "#6f9dc6")],
                "darkcolor" : [("focus", "#6f9dc6")]
            }
        },
        
        "TCombobox" : {
            "configure" : {
                "padding" : [scale(4),scale(2),scale(2),scale(2)],
                "insertwidth" : scale(1),
            },
            "map" : {
                "background" : [("active", lighter),
                                ("pressed", lighter)],
                "fieldbackground" : [("readonly", "focus", selectbg),
                                     ("readonly", frame)],
                "foreground" : [("readonly", "focus", selectfg)],
                "arrowcolor" : [("disabled", disabledfg)]
            }
        },
        
        "ComboboxPopdownFrame" : {
            "configure" : {
                "relief" : "solid",
                "borderwidth" : scale(1)
            }
        },
        
        "TSpinbox" : {
            "configure" : {
                "arrowsize" : scale(10),
                "padding" : [scale(2), 0, scale(10), 0],
            },
            "map" : {
                "background" : [("readonly", frame)],
                "arrowcolor" : [("disabled", disabledfg)]
            }
        },
        
        "TNotebook.Tab" : {
            "configure" : {
                "padding" : [scale(6), scale(2), scale(6), scale(2)]
            },
            "map" : {
                "padding" : [("selected", [scale(6), scale(4), scale(6), scale(4)])],
                "background" : [("selected", frame), 
                                ("",        darker)],
                "lightcolor" : [("selected", lighter), 
                                ("",        dark)], 
            }
        },
        
        "Treeview" : {
            "configure" : {
                "background" : window,
            },
            "map" : {
                "background" : [("disabled", frame),
                                ("!disabled", "!selected", window),
                                ("selected", selectbg)],
                "foreground" : [("disabled", disabledfg),
                                ("!disabled", "!selected", defaultfg),
                                ("selected", selectfg)]        
            }
        },
        
        # Treeview heading
        "Heading" : {
            "configure" : {
                "font" : "TkHeadingFont",
                "relief" : "raised",
                "padding" : [scale(3), scale(3), scale(3), scale(3)]
            }
        },
        
        "TLabelframe" : {
            "configure" : {
                "labeloutside" : True,
                "labelmargins" : [0, 0, 0, scale(4)]
            }
        },
        
        "TProgressbar" : {
            "configure" : {
                "background" : frame
            }
        },
        
        "Sash" : {
            "configure" : {
                "sashthickness" : scale(6),
                "gripcount" : 10
            }
        },
    }
    

def xpnative() -> BasicUiThemeSettings:
    # Transcribed from https://github.com/tcltk/tk/blob/master/library/ttk/xpTheme.tcl
    return {
        "." : {
            "configure" : {
                "background"        : "SystemButtonFace",
                "foreground"        : "SystemWindowText",
                "selectbackground"  : "SystemHighlightText",
                "selectforeground"  : "SystemHighlight",
                "font" : "TkDefaultFont",
            },
            
            "map" : {
                "foreground" : [("disabled", "SystemGrayText")],
            },
        },
            
        "TButton" : {
            "configure" : {
                "anchor" : "center", 
                "width" : scale(11), 
                "padding" : [scale(1), scale(1)], 
            },
        },
        
        "Toolbutton" : {
            "configure" : {
                "padding" :[scale(4), scale(4)],
            },
        },
        
        "TCheckbutton" : {
            "configure" : {
                "padding" :  scale(2),
            },
        },
        
        # TRadiobutton has same style as TCheckbutton
        "TRadiobutton" : {
            "configure" : {
                "padding" :  scale(2),
            },
        },
        
        "TMenubutton" : {
            "configure" : {
                "padding" : [scale(8), scale(4)],
            }
        },
        
        "TEntry" : {
            "configure" : {
                "padding" : [scale(2), scale(2), scale(2), scale(4)],
            },
            "map" : {
                "selectbackground" : [("!focus", "SystemWindow")],
                "selectforeground" : [("!focus", "SystemWindowText")]
            }
        },
        
        "TCombobox" : {
            "configure" : {
                "padding" : scale(2),
            },
            "map" : {
                "selectbackground" : [("!focus", "SystemWindow")],
                "selectforeground" : [("!focus", "SystemWindowText")],
                "foreground" : [("disabled", "SystemGrayText"),
                                ("readonly", "focus", "SystemHighlightText")],
                "focusfill" : [("readonly", "focus", "SystemHighlight")]
            }
        },
        
        "ComboboxPopdownFrame" : {
            "configure" : {
                "relief" : "solid",
                "borderwidth" : scale(1)
            }
        },
        
        "TSpinbox" : {
            "configure" : {
                "padding" : [scale(2), 0, scale(14), 0],
            },
            "map" : {
                "selectbackground" : [("!focus", "SystemWindow")],
                "selectforeground" : [("!focus", "SystemWindowText")],
            }
        },
        
        "TNotebook" : {
            "configure" : {
                "tabmargins" : [scale(2), scale(2), scale(2), 0]
            }
        },
        
        "TNotebook.Tab" : {
            "map" : {
                "expand" : [("selected", [scale(2), scale(2), scale(2), scale(2)])],
            }
        },
        
        "Treeview" : {
            "configure" : {
                "background" : "SystemWindow",
            },
            "map" : {
                "background" : [("disabled", "SystemButtonFace"),
                                ("!disabled", "!selected", "SystemWindow"),
                                ("selected", "SystemHighlight")],
                "foreground" : [("disabled", "SystemGrayText"),
                                ("!disabled", "!selected", "SystemWindowText"),
                                ("selected", "SystemHighlightText")]        
            }
        },
        
        
        "Heading" : { # Treeview heading
            "configure" : {
                "font" : "TkHeadingFont",
                "relief" : "raised"
            }
        },
        
        "TLabelframe.Label" : {
            "configure" : {
                "foreground" : "#0046d5",
            }
        },
    }

def aqua() -> BasicUiThemeSettings:
    # https://github.com/tcltk/tk/blob/master/library/ttk/aquaTheme.tcl
    return {
        "." : {
            "configure" : {
                "font" : "TkDefaultFont",
                "background" : "systemWindowBody",
                "foreground" : "systemModelessDialogActiveText",
                "selectbackground" : "systemHighlight",
                "selectforeground" : "systemModelessDialogActiveText",
                "selectborderwidth" : 0,
                "insertwidth" : 1,
                "stipple" : "",
            },
            "map" : {
                "foreground" : [("disabled", "systemModelessDialogInactiveText"),
                                ("background", "systemModelessDialogInactiveText")],
                "selectbackground" : [("background", "systemHighlightSecondary"),
                                      ("!focus", "systemHighlightSecondary")],
                "selectforeground" : [("background", "systemModelessDialogInactiveText"),
                                      ("!focus", "systemDialogActiveText")],
            },
        },
        "TButton" : {
            "configure" : {
                "anchor" : "center",
                "width" : "6"
            },
        },
        "Toolbutton" : {
            "configure" : {
                "padding" : 4,
            },
        },
        "TNotebook" : {
            "configure" : {
                "tabmargins" : [10, 0],
                "tabposition" : "n",
                "padding" : [18, 8, 18, 17],
            },
        },
        "TNotebook.Tab" : {
            "configure" : {
                "padding" : [12, 3, 12, 2],
            },
        },
        "TCombobox" : {
            "configure" : {
                "postoffset" : [5, -2, -10, 0]
            },
        },
        "Heading" : {
            "configure" : {
                "font" : "TkHeadingFont",
            },
        },
        "Treeview" : {
            "configure" : {
                "rowheight" : 18,
                "background" : "white"
            },
            "map" : {
                "background" : [("disabled", "systemDialogBackgroundInactive"),
                                ("!disabled", "!selected", "systemWindowBody"),
                                ("selected", "background", "systemHighlightSecondary"),
                                ("selected", "systemHighlight")],
                "foreground" : [("disabled", "systemModelessDialogInactiveText"),
                                ("!disabled", "!selected", "black"),
                                ("selected", "systemModelessDialogActiveText")],
            },
        },
        "TProgressbar" : {
            "configure" : {
                "period" : 100,
                "maxphase" : 255
            },
        },
        "Labelframe" : {
            "configure" : {
                "labeloutside" : True,
                "labelmargins" : [14, 0, 14, 4]
            },
        },
    }

def windows() -> CompoundUiThemeSettings:
    return [
        xpnative(),
        _treeview_settings(),
        _menubutton_settings(),
        _paned_window_settings(),
        _menu_settings(),
        _text_settings(),
        _label_settings(),
        _button_notebook_settings(),
        {
            "TNotebook" : {
                "configure" : {
                    # With tabmargins I can get a gray line below tab, which separates
                    # tab content from label
                    "tabmargins" : [scale(2), scale(2), scale(2), scale(2)]
                }
            },
            "Tab" : {
                "configure" : {
                    "padding" : [scale(3),scale(1),scale(3),0]
                }
            },
            "ButtonNotebook.TNotebook.Tab" : {
                "configure" : {
                    "padding" : (scale(4),scale(1),scale(1),0)
                }
            },
            
            "TCombobox" : {
                "map" : {
                    "selectbackground"  : [("readonly", "!focus", "SystemWindow"),
                                           ("readonly", "focus", "SystemHighlight")],
                    "selectforeground"  : [("readonly", "!focus", "SystemWindowText"),
                                           ("readonly", "focus", "SystemHighlightText")],
                }
            },
            
            "Listbox" : {
                "configure" : {
                    "background" : "SystemWindow",
                    "foreground" : "SystemWindowText",
                    "disabledforeground" : "SystemGrayText",
                    "highlightbackground" : "SystemActiveBorder",
                    "highlightcolor" : "SystemActiveBorder",
                    "highlightthickness" : scale(1),
                },
            },
            "ViewBody.TFrame" : {
                "configure" : {
                    "background" : "SystemButtonFace", # to create the fine line below toolbar
                }
            },
            
            "ViewToolbar.TFrame" : {
                "configure" : {
                    "background" : "SystemWindow",
                }
            },
            
            "ViewToolbar.Toolbutton" : {
                "configure" : {
                    "background" : "SystemWindow",
                }
            },
            
            "ViewTab.TLabel" : {
                "configure" : {
                    "background" : "SystemWindow",
                    "padding" : [scale(5), 0],
                },
            },
            
            "ViewToolbar.TLabel" : {
                "configure" : {
                    "background" : "SystemWindow",
                    "padding" : [scale(5), 0],
                },
            },
            
            "Active.ViewTab.TLabel" : {
                "configure" : {
                    #"font" : "BoldTkDefaultFont",
                    "relief" : "sunken",
                    "borderwidth" : scale(1),
                }
            },
            
            "Inactive.ViewTab.TLabel" : {
                "configure" : {
                    "font" : "UnderlineTkDefaultFont",
                },
            },
        }
    ]

def enhanced_clam() -> CompoundUiThemeSettings:
    return [
        clam(),
        _treeview_settings(),
        _menubutton_settings(),
        _paned_window_settings(),
        _menu_settings(),
        _text_settings(),
        _label_settings(),
        _button_notebook_settings(),
        {
            "ButtonNotebook.Tab" : {
                "configure" : {
                    "padding" : (scale(6),scale(4),scale(2),scale(3))
                }
            },
            
            "TScrollbar" : {
                "configure" : {
                    "gripcount" : 0,
                    "arrowsize" : scale(14),
                    #"arrowcolor" : "DarkGray"
                    #"width" : 99 # no effect
                }
            },
            
            "TCombobox" : {
                "configure" : {
                    "arrowsize" : scale(14),
                },
                "map" : {
                    "selectbackground"  : [("readonly", "!focus", "#dcdad5")],
                    "selectforeground"  : [("readonly", "!focus", "#000000")],
                }
            },
                
            "TCheckbutton" : {
                "configure" : {
                    "indicatorsize" : scale(12)
                }
            },
            
            "TRadiobutton" : {
                "configure" : {
                    "indicatorsize" : scale(12)
                }
            },
            
            "Listbox" : {
                "configure" : {
                    "background" : "white",
                    "foreground" : "black",
                    "disabledforeground" : "#999999",
                    "highlightbackground" : "#4a6984",
                    "highlightcolor" : "#4a6984",
                    "highlightthickness" : scale(1),
                },
            },
        }
    ]

def enhanced_aqua() -> CompoundUiThemeSettings:
    return [
        _treeview_settings(),
        _menubutton_settings(),
        #_paned_window_settings(),
        _menu_settings(),
        {
            "TPanedWindow" : {
                "configure" : {
                    "background" : "systemDialogBackgroundActive",
                }
            },
            "TFrame" : {
                "configure" : {
                    "background" : "systemDialogBackgroundActive",
                }
            },
            "Tab" : {
                "map" : {
                    "foreground" : [("selected", "white")]
                }
            },
        }
    ]



def load_plugin() -> None:
    from tkinter import ttk
    original_themes = ttk.Style().theme_names() 
    
    # load all base themes
    for name in original_themes:
        settings = {} # type: Union[Dict, Callable[[], Dict]]
        if name == "clam":
            settings = clam
        elif name == "xpnative":
            settings = xpnative
        elif name == "aqua":
            settings = aqua
             
        get_workbench().add_ui_theme(name, None, settings)
    
    get_workbench().add_ui_theme("Enhanced Clam", "clam", enhanced_clam, 
                                 {"tab-close" : "tab-close-clam",
                                  "tab-close-active" : "tab-close-active-clam"})
    
    if "xpnative" in original_themes:
        get_workbench().add_ui_theme("Windows", "xpnative", windows)
    
    if "aqua" in original_themes:
        get_workbench().add_ui_theme("Kind of Aqua", "aqua", enhanced_aqua)
    

    if "Windows" in get_workbench().get_usable_ui_theme_names():
        get_workbench().set_default("view.ui_theme", "Windows")
        
    elif "Enhanced Clam" in get_workbench().get_usable_ui_theme_names():
        get_workbench().set_default("view.ui_theme", "Enhanced Clam")



"""
NB! Stippling doesn't work on mac: 
http://wiki.tcl.tk/44444
http://rkeene.org/projects/tcl/tk.fossil/tkthistory/2954673
"""
import logging
import os.path
from tkinter import font

import thonny
from thonny import get_workbench, jedi_utils
from thonny.codeview import get_syntax_options_for_tag

python_tree = jedi_utils.import_python_tree()

def create_bitmap_file(width, height, predicate, name):
    
    cache_dir = os.path.join(thonny.THONNY_USER_DIR, "image_cache")
    name = "%s_%d_%d.xbm" % (name, width, height)
    filename = os.path.join(cache_dir, name)
    
    #if os.path.exists(filename):
    #    return filename
    
    hex_lines = []
    
    if width % 8 == 0:
        row_size = width
    else:
        # need to pad row size so that it is multiple of 8
        row_size = width + 8 - (width % 8)
    
    for y in range(height):
        byte_hexes = []
        for byte_index in range(row_size // 8):
            byte = 0
            for bit_index in range(7,-1,-1):
                x = byte_index * 8 + bit_index
                
                byte <<= 1
                if predicate(x, y):
                    byte |= 1
                    
            byte_hexes.append(format(byte, "#04x"))
        hex_lines.append(",".join(byte_hexes))
                
    data = (
        "#define im_width %d\n" % width
        + "#define im_height %d\n" % height
        + "static char im_bits[] = {\n"
        + "%s\n" % ",\n".join(hex_lines)
        + "};"
    )
    
    os.makedirs(cache_dir, exist_ok=True)
    with open(filename, "w") as fp:
        fp.write(data)
    return filename
    


def configure_text(text):
    spacing1 = 2
    spacing3 = 3
    text_font = text["font"]
    text.configure(spacing1=spacing1, spacing3=spacing3)
    text.master._gutter.configure(spacing1=spacing1, spacing3=spacing3)
    if isinstance(text_font, str):
        text_font = font.nametofont(text_font)
    
    indent_width = text_font.measure("    ")
    bbox = text.bbox("1.0")
    if bbox is None or bbox[3] < 5:
        # text not ready yet
        # TODO: Text in Tk 8.6 has sync method
        return False
    
    line_height = bbox[3] + spacing1 + spacing3
    
    print(indent_width, line_height)
    
    def ver(x: int, y: int, top: bool, bottom: bool) -> bool:
        # tells where to show pixels in vertical border of the statement
        # It would be convenient if tiling started from the start of 
        # 1st char, but it is offset a bit
        # In order to make computation easier, I'm offsetting x as well
        x = (x-5) % indent_width
        
        stripe_width = 8
        gap = 3
        left = indent_width-stripe_width - gap
        
        return (left <= x < left+stripe_width 
                or top and y == 0 and x >= left
                or bottom and y == line_height-1 and x >= left
        )
        
    def hor(x: int, y: int, top: bool, bottom: bool) -> bool:
        # tells where to show pixels in statement line
        return (
            top and y == 0
            or bottom and y == line_height-1
        )
    
    color = get_syntax_options_for_tag("GUTTER").get("background", "gray")
    for orient, base_predicate in [("hor", hor), ("ver", ver)]:
        for top in [False, True]:
            for bottom in [False, True]:
                def predicate(x, y, 
                              # need to make base_predicate, top and bottom local
                              base_predicate=base_predicate,
                              top=top,
                              bottom=bottom):
                    return base_predicate(x, y, top, bottom)
                    
                tag_name = "%s_%s_%s" % (orient, top, bottom)
                bitmap_path = create_bitmap_file(indent_width, line_height, 
                                                 predicate, tag_name)
                text.tag_configure(tag_name, background=color,
                                   bgstipple="@" + bitmap_path)
    
    return True

def print_tree(node, level=0):
    indent = "  " * level
    #if (isinstance(node, python_tree.PythonNode) and node.type == "sim"
    if (node.type in ("simple_stmt",)
        or isinstance(node, python_tree.Flow)
        ):
        print(indent, node.type, node.start_pos, node.end_pos)
        
    if hasattr(node, "children"):
        for child in node.children:
            print_tree(child, level+1)

def clear_tags(text):
    for pos in ["ver", "hor"]:
        for top in [True, False]:
            for bottom in [True, False]:
                text.tag_remove("%s_%s_%s" % (pos, top, bottom), "1.0", "end")

def add_tags(text):
    source = text.get("1.0", "end")
    clear_tags(text)
    tree = jedi_utils.parse_source(source)
    
    print_tree(tree)
    last_line = 0
    last_col = 0
    
    def tag_tree(node):
        nonlocal last_line, last_col
        
        if (node.type == "simple_stmt"
            or isinstance(node, (python_tree.Flow, python_tree.Scope))):
            
            start_line, start_col = node.start_pos
            end_line, end_col = node.end_pos
            
            # Before dealing with this node,
            # handle the case, where last vertical tag was meant for 
            # same column, but there were empty or comment lines between
            if start_col == last_col:
                for i in range(last_line+1, start_line):
                    # NB! tag not visible when logically empty line
                    # doesn't have indent prefix
                    text.tag_add("ver_False_False",
                                "%d.%d" % (i, last_col-1),
                                "%d.%d" % (i, last_col))
                    print("ver_False_False",
                                "%d.%d" % (i, last_col-1),
                                "%d.%d" % (i, last_col))
                    
            
            print(node)
            
            # usually end_col is 0
            # exceptions: several statements on the same line (semicoloned statements)
            # also unclosed parens in if-header  
            for lineno in range(start_line, 
                                end_line if end_col == 0 else end_line+1):
            
                top = (lineno == start_line and lineno > 1)
                bottom = False # start_line == end_line-1
            
                # horizontal line (only for first or last line)
                if top or bottom:
                    text.tag_add("hor_%s_%s" % (top, bottom), 
                                 "%d.%d" % (lineno, start_col), 
                                 "%d.%d" % (lineno+1 if end_col == 0 else lineno,
                                            0))
                    
                    print("hor_%s_%s" % (top, bottom), 
                                 "%d.%d" % (lineno, start_col), 
                                 "%d.%d" % (lineno+1, 0))
                
                # vertical line (only for indented statements)
                # Note that I'm using start col for all lines
                # (statement's indent shouldn't decrease in continuation lines)
                if start_col > 0:
                    text.tag_add("ver_%s_%s" % (top, bottom),
                                 "%d.%d" % (lineno, start_col-1),
                                 "%d.%d" % (lineno, start_col))
                    print("ver_%s_%s" % (top, bottom),
                                 "%d.%d" % (lineno, start_col-1),
                                 "%d.%d" % (lineno, start_col))
                    
                    last_line = lineno
                    last_col = start_col
        
        # Recurse
        if node.type != "simple_stmt" and hasattr(node, "children"):
            for child in node.children:
                tag_tree(child)
    
    tag_tree(tree)

def handle_editor_event(event):
    configure_and_add_tags(event.editor.get_text_widget())

def handle_events(event):
    if hasattr(event, "text_widget"):
        text = event.text_widget
    else:
        text = event.widget
    
    configure_and_add_tags(text)

def configure_and_add_tags(text):
    if not getattr(text, "structure_tags_configured", False):
        try:
            if configure_text(text):
                text.structure_tags_configured = True
            else:
                text.after(500, lambda: configure_and_add_tags(text))
                return
        except Exception:
            logging.exception("Problem with defining structure tags")
            return
    
    add_tags(text)
        

def _load_plugin() -> None:
    wb = get_workbench() 

    wb.set_default("view.program_structure", False)
    wb.bind("Save", handle_editor_event, True)
    wb.bind("Open", handle_editor_event, True)
    wb.bind_class("CodeViewText", "<<TextChange>>", handle_events, True)




class ArithmeticError:
    ''

class AssertionError:
    ''

class AttributeError:
    ''

class BaseException:
    ''

class EOFError:
    ''
Ellipsis = None

class Exception:
    ''

class GeneratorExit:
    ''

class ImportError:
    ''

class IndentationError:
    ''

class IndexError:
    ''

class KeyError:
    ''

class KeyboardInterrupt:
    ''

class LookupError:
    ''

class MemoryError:
    ''

class NameError:
    ''

class NotImplementedError:
    ''

class OSError:
    ''

class OverflowError:
    ''

class RuntimeError:
    ''

class StopAsyncIteration:
    ''

class StopIteration:
    ''

class SyntaxError:
    ''

class SystemExit:
    ''

class TypeError:
    ''

class UnicodeError:
    ''

class ValueError:
    ''

class ViperTypeError:
    ''

class ZeroDivisionError:
    ''
def abs():
    pass

def all():
    pass

def any():
    pass

def bin():
    pass


class bool:
    ''

class bytearray:
    ''
    def append():
        pass

    def extend():
        pass


class bytes:
    ''
    def center():
        pass

    def count():
        pass

    def decode():
        pass

    def endswith():
        pass

    def find():
        pass

    def format():
        pass

    def index():
        pass

    def isalpha():
        pass

    def isdigit():
        pass

    def islower():
        pass

    def isspace():
        pass

    def isupper():
        pass

    def join():
        pass

    def lower():
        pass

    def lstrip():
        pass

    def partition():
        pass

    def replace():
        pass

    def rfind():
        pass

    def rindex():
        pass

    def rpartition():
        pass

    def rsplit():
        pass

    def rstrip():
        pass

    def split():
        pass

    def splitlines():
        pass

    def startswith():
        pass

    def strip():
        pass

    def upper():
        pass

def callable():
    pass

def chr():
    pass


class classmethod:
    ''
def compile():
    pass


class complex:
    ''
def delattr():
    pass


class dict:
    ''
    def clear():
        pass

    def copy():
        pass

    def fromkeys():
        pass

    def get():
        pass

    def items():
        pass

    def keys():
        pass

    def pop():
        pass

    def popitem():
        pass

    def setdefault():
        pass

    def update():
        pass

    def values():
        pass

def dir():
    pass

def divmod():
    pass


class enumerate:
    ''
def eval():
    pass

def exec():
    pass

def execfile():
    pass


class filter:
    ''

class float:
    ''

class frozenset:
    ''
    def copy():
        pass

    def difference():
        pass

    def intersection():
        pass

    def isdisjoint():
        pass

    def issubset():
        pass

    def issuperset():
        pass

    def symmetric_difference():
        pass

    def union():
        pass

def getattr():
    pass

def globals():
    pass

def hasattr():
    pass

def hash():
    pass

def help():
    pass

def hex():
    pass

def id():
    pass

def input():
    pass


class int:
    ''
    def from_bytes():
        pass

    def to_bytes():
        pass

def isinstance():
    pass

def issubclass():
    pass

def iter():
    pass

def len():
    pass


class list:
    ''
    def append():
        pass

    def clear():
        pass

    def copy():
        pass

    def count():
        pass

    def extend():
        pass

    def index():
        pass

    def insert():
        pass

    def pop():
        pass

    def remove():
        pass

    def reverse():
        pass

    def sort():
        pass

def locals():
    pass


class map:
    ''
def max():
    pass


class memoryview:
    ''
def min():
    pass

def next():
    pass


class object:
    ''
def oct():
    pass

def open():
    pass

def ord():
    pass

def pow():
    pass

def print():
    pass


class property:
    ''
    def deleter():
        pass

    def getter():
        pass

    def setter():
        pass


class range:
    ''
def repr():
    pass


class reversed:
    ''
def round():
    pass


class set:
    ''
    def add():
        pass

    def clear():
        pass

    def copy():
        pass

    def difference():
        pass

    def difference_update():
        pass

    def discard():
        pass

    def intersection():
        pass

    def intersection_update():
        pass

    def isdisjoint():
        pass

    def issubset():
        pass

    def issuperset():
        pass

    def pop():
        pass

    def remove():
        pass

    def symmetric_difference():
        pass

    def symmetric_difference_update():
        pass

    def union():
        pass

    def update():
        pass

def setattr():
    pass


class slice:
    ''
def sorted():
    pass


class staticmethod:
    ''

class str:
    ''
    def center():
        pass

    def count():
        pass

    def encode():
        pass

    def endswith():
        pass

    def find():
        pass

    def format():
        pass

    def index():
        pass

    def isalpha():
        pass

    def isdigit():
        pass

    def islower():
        pass

    def isspace():
        pass

    def isupper():
        pass

    def join():
        pass

    def lower():
        pass

    def lstrip():
        pass

    def partition():
        pass

    def replace():
        pass

    def rfind():
        pass

    def rindex():
        pass

    def rpartition():
        pass

    def rsplit():
        pass

    def rstrip():
        pass

    def split():
        pass

    def splitlines():
        pass

    def startswith():
        pass

    def strip():
        pass

    def upper():
        pass

def sum():
    pass


class super:
    ''

class tuple:
    ''
    def count():
        pass

    def index():
        pass


class type:
    ''

class zip:
    ''



# -*- coding: utf-8 -*-

import datetime
import platform
import tkinter as tk
import tkinter.font
import webbrowser
from tkinter import ttk

import thonny
from thonny import get_workbench, misc_utils, ui_utils
from thonny.misc_utils import get_python_version_string


class AboutDialog(tk.Toplevel):
    def __init__(self, master):
        tk.Toplevel.__init__(self, master)
        

        main_frame = ttk.Frame(self)
        main_frame.grid(sticky=tk.NSEW, ipadx=15, ipady=15)
        main_frame.rowconfigure(0, weight=1)
        main_frame.columnconfigure(0, weight=1)

        self.title("About Thonny")
        if misc_utils.running_on_mac_os():
            self.configure(background="systemSheetBackground")
        self.resizable(height=tk.FALSE, width=tk.FALSE)
        self.protocol("WM_DELETE_WINDOW", self._ok)
        
        
        #bg_frame = ttk.Frame(self) # gives proper color in aqua
        #bg_frame.grid()
        
        heading_font = tkinter.font.nametofont("TkHeadingFont").copy()
        heading_font.configure(size=19, weight="bold")
        heading_label = ttk.Label(main_frame, 
                                  text="Thonny " + thonny.get_version(),
                                  font=heading_font)
        heading_label.grid()
        
        
        url = "https://thonny.org"
        url_font = tkinter.font.nametofont("TkDefaultFont").copy()
        url_font.configure(underline=1)
        url_label = ttk.Label(main_frame, text=url, style="Url.TLabel",
                              cursor="hand2",
                              font=url_font,)
        url_label.grid()
        url_label.bind("<Button-1>", lambda _:webbrowser.open(url))
        
        if platform.system() == "Linux":
            try:
                import distro # distro don't need to be installed
                system_desc = distro.name(True)
            except ImportError:
                system_desc = "Linux"
                
            if "32" not in system_desc and "64" not in system_desc:
                system_desc += " " + self.get_os_word_size_guess()
        else:
            system_desc = (platform.system() 
                        + " " + platform.release()  
                        + " " + self.get_os_word_size_guess())
        
        platform_label = ttk.Label(main_frame, justify=tk.CENTER, 
                                   text= system_desc + "\n"
                                        + "Python " + get_python_version_string() 
                                        + "Tk " + ui_utils.get_tk_version_str())
        platform_label.grid(pady=20)
        
        credits_label = ttk.Label(main_frame, 
                                  text="Made in\n"
                                  + "University of Tartu, Estonia,\n"
                                  + "with the help from\n"
                                  + "open-source community\n"
                                  + "and "
                                  + "Raspberry Pi Foundation",
                              style="Url.TLabel",
                              cursor="hand2",
                              font=url_font,
                              justify="center")
        credits_label.grid()
        credits_label.bind("<Button-1>", lambda _:webbrowser.open("https://bitbucket.org/plas/thonny/src/master/CREDITS.rst"))
        
        license_font = tkinter.font.nametofont("TkDefaultFont").copy()
        license_font.configure(size=7)
        license_label = ttk.Label(main_frame,
                                  text="Copyright (©) "
                                  + str(datetime.datetime.now().year)
                                  + " Aivar Annamaa\n"
                                  + "This program comes with\n"
                                  + "ABSOLUTELY NO WARRANTY!\n"
                                  + "It is free software, and you are welcome to\n"
                                  + "redistribute it under certain conditions, see\n"
                                  + "https://opensource.org/licenses/MIT\n"
                                  + "for details",
                                  justify=tk.CENTER, font=license_font)
        license_label.grid(pady=20)
        
        
        ok_button = ttk.Button(main_frame, text="OK", command=self._ok, default="active")
        ok_button.grid(pady=(0,15))
        ok_button.focus_set()
        
        self.bind('<Return>', self._ok, True) 
        self.bind('<Escape>', self._ok, True)
        
        
    def _ok(self, event=None):
        self.destroy()
    
    def get_os_word_size_guess(self):
        if "32" in platform.machine() and "64" not in platform.machine():
            return "(32-bit)"
        elif "64" in platform.machine() and "32" not in platform.machine():
            return "(64-bit)"
        else:
            return ""


def load_plugin() -> None:
    def open_about(*args):
        ui_utils.show_dialog(AboutDialog(get_workbench()))
    
    def open_url(url):
        # webbrowser.open returns bool, but add_command expects None
        webbrowser.open(url)
        
    get_workbench().add_command("changelog", "help", "Version history",
                                lambda: open_url("https://bitbucket.org/plas/thonny/src/master/CHANGELOG.rst"), group=60)
    get_workbench().add_command("issues", "help", "Report problems",
                                lambda: open_url("https://bitbucket.org/plas/thonny/issues/new"), group=60)
    get_workbench().add_command("about", "help", "About Thonny", open_about, group=61)
    
    # For Mac
    get_workbench().createcommand("tkAboutDialog", open_about)



import tkinter as tk

from thonny import get_workbench
from thonny.common import TextRange
from thonny.ui_utils import select_sequence

BLOCK_COMMENT_PREFIX = "#"

def _get_focused_writable_text():
    widget = get_workbench().focus_get()
    # In Ubuntu when moving from one menu to another, this may give None when text is actually focused
    if (isinstance(widget, tk.Text) 
        and (not hasattr(widget, "is_read_only") or not widget.is_read_only())):
        return widget 
    else:
        return None

def _writable_text_is_focused():
    return _get_focused_writable_text() is not None 


def _selection_is_line_commented(text):
    sel_range = _get_focused_code_range(text)
    
    for lineno in range(sel_range.lineno, sel_range.end_lineno+1):
        line = text.get(str(lineno) + '.0', str(lineno) + '.end')
        if not line.startswith(BLOCK_COMMENT_PREFIX):
            return False
    
    return True
    
def _select_lines(text, first_line, last_line):
    text.tag_remove("sel", "1.0", tk.END)
    text.tag_add("sel", 
                      str(first_line) + ".0",
                      str(last_line) + ".end")

def _toggle_selection_comment(text):
    if _selection_is_line_commented(text):
        _uncomment_selection(text)
    else:
        _comment_selection(text)
    

def _comment_selection(text):
    """Adds ## in front of all selected lines if any lines are selected, 
    or just the current line otherwise"""
    
    sel_range = _get_focused_code_range(text)

    for lineno in range(sel_range.lineno, sel_range.end_lineno+1):
        text.insert(str(lineno) + '.0', BLOCK_COMMENT_PREFIX)
    
    if sel_range.end_lineno > sel_range.lineno:
        _select_lines(text, sel_range.lineno, sel_range.end_lineno)
    
    text.edit_separator()
        

def _uncomment_selection(text):
    sel_range = _get_focused_code_range(text)
    
    for lineno in range(sel_range.lineno, sel_range.end_lineno+1):
        line = text.get(str(lineno) + '.0', str(lineno) + '.end')
        if line.startswith(BLOCK_COMMENT_PREFIX):
            text.delete(str(lineno) + ".0",
                             str(lineno) + "." + str(len(BLOCK_COMMENT_PREFIX)))

def _get_focused_code_range(text):
    if len(text.tag_ranges("sel")) > 0:
        lineno, col_offset = map(int, text.index(tk.SEL_FIRST).split("."))
        end_lineno, end_col_offset = map(int, text.index(tk.SEL_LAST).split("."))
        
        if end_lineno > lineno and end_col_offset == 0:
            # SelectAll includes nonexisting extra line
            end_lineno -= 1
            end_col_offset = int(text.index(str(end_lineno) + ".end").split(".")[1])
    else:
        lineno, col_offset = map(int, text.index(tk.INSERT).split("."))
        end_lineno, end_col_offset = lineno, col_offset
        
    return TextRange(lineno, col_offset, end_lineno, end_col_offset)


def _cmd_toggle_selection_comment():
    text = _get_focused_writable_text()
    if text is not None: 
        _toggle_selection_comment(text)
        
def _cmd_comment_selection():
    text = _get_focused_writable_text()
    if text is not None: 
        _comment_selection(text)

def _cmd_uncomment_selection():
    text = _get_focused_writable_text()
    if text is not None: 
        _uncomment_selection(text)



def load_plugin() -> None:
    
    get_workbench().add_command("toggle_comment", "edit", "Toggle comment",
        _cmd_toggle_selection_comment,
        default_sequence=select_sequence("<Control-Key-3>", "<Command-Key-3>"),
        tester=_writable_text_is_focused,
        group=50)
    
    get_workbench().add_command("comment_selection", "edit", "Comment out",
        _cmd_comment_selection,
        default_sequence="<Alt-Key-3>",
        tester=_writable_text_is_focused,
        group=50)
    
    get_workbench().add_command("uncomment_selection", "edit", "Uncomment",
        _cmd_uncomment_selection,
        default_sequence="<Alt-Key-4>",
        tester=_writable_text_is_focused,
        group=50)




class VfsFat:
    ''
    def chdir():
        pass

    def getcwd():
        pass

    def ilistdir():
        pass

    def mkdir():
        pass

    def mkfs():
        pass

    def mount():
        pass

    def open():
        pass

    def remove():
        pass

    def rename():
        pass

    def rmdir():
        pass

    def stat():
        pass

    def statvfs():
        pass

    def umount():
        pass

def chdir():
    pass

def dupterm():
    pass

def getcwd():
    pass

def ilistdir():
    pass

def listdir():
    pass

def mkdir():
    pass

def mount():
    pass

def remove():
    pass

def rename():
    pass

def rmdir():
    pass

sep = '/'
def stat():
    pass

def statvfs():
    pass

def sync():
    pass

def umount():
    pass

def uname():
    pass

def unlink():
    pass



# -*- coding: utf-8 -*-
"""Adds commands for opening certain Thonny folders"""

from thonny import THONNY_USER_DIR, get_workbench
from thonny.ui_utils import open_path_in_system_file_manager


def load_plugin() -> None:
    def cmd_open_data_dir():
        open_path_in_system_file_manager(THONNY_USER_DIR)
        
    def cmd_open_program_dir():
        open_path_in_system_file_manager(get_workbench().get_package_dir())
        
    get_workbench().add_command("open_program_dir", "tools", "Open Thonny program folder...",
                                cmd_open_program_dir, group=110)
    get_workbench().add_command("open_data_dir", "tools", "Open Thonny data folder...",
                                cmd_open_data_dir, group=110)



import textwrap
import tkinter as tk
from tkinter import font as tk_font
from tkinter import ttk

from thonny import get_workbench
from thonny.codeview import CodeView
from thonny.config_ui import ConfigurationPage
from thonny.ui_utils import create_string_var


class ThemeAndFontConfigurationPage(ConfigurationPage):
    
    def __init__(self, master):
        
        self._original_family = get_workbench().get_option("view.editor_font_family")
        self._original_size = get_workbench().get_option("view.editor_font_size")
        self._original_ui_theme = get_workbench().get_option("view.ui_theme")
        self._original_syntax_theme = get_workbench().get_option("view.syntax_theme")
        
        ConfigurationPage.__init__(self, master)
        
        self._family_variable = create_string_var(self._original_family,
            modification_listener=self._update_appearance)
        
        self._size_variable = create_string_var(self._original_size,
            modification_listener=self._update_appearance)
        
        self._ui_theme_variable = create_string_var(self._original_ui_theme,
            modification_listener=self._update_appearance)
        
        self._syntax_theme_variable = create_string_var(self._original_syntax_theme,
            modification_listener=self._update_appearance)
        
        ttk.Label(self, text="UI theme").grid(row=1, column=0, sticky="w", pady=(10, 0))
        self._ui_theme_combo = ttk.Combobox(self,
                                        exportselection=False,
                                        textvariable=self._ui_theme_variable,
                                        state='readonly',
                                        height=15,
                                        values=get_workbench().get_usable_ui_theme_names())
        self._ui_theme_combo.grid(row=2, column=0, sticky="nsew", padx=(0, 10))
        
        ttk.Label(self, text="Syntax theme").grid(row=1, column=1, sticky="w", pady=(10, 0))
        self._syntax_theme_combo = ttk.Combobox(self,
                                        exportselection=False,
                                        textvariable=self._syntax_theme_variable,
                                        state='readonly',
                                        height=15,
                                        values=get_workbench().get_syntax_theme_names())
        self._syntax_theme_combo.grid(row=2, column=1, sticky="nsew", padx=(0, 10))
        
        
        ttk.Label(self, text="Editor font").grid(row=1, column=2, sticky="w", pady=(10, 0))
        self._family_combo = ttk.Combobox(self,
                                          exportselection=False,
                                          state='readonly',
                                          height=15,
                                          textvariable=self._family_variable,
                                          values=self._get_families_to_show())
        self._family_combo.grid(row=2, column=2, sticky=tk.NSEW, padx=(0,10))
        
        ttk.Label(self, text="Size").grid(row=1, column=3, sticky="w", pady=(10, 0))
        self._size_combo = ttk.Combobox(self, width=4,
                                        exportselection=False,
                                        textvariable=self._size_variable,
                                        state='readonly',
                                        height=15,
                                      values=[str(x) for x in range(3,73)])
        self._size_combo.grid(row=2, column=3, sticky="nsew")
        
        
        ttk.Label(self, text="Editor preview").grid(row=3, column=0, sticky="w", 
                                                    pady=(10,0), columnspan=4)
        self._preview_codeview = CodeView(self,
                                height=10,
                                font="EditorFont",
                                #relief="sunken",
                                #borderwidth=1,
                                )
        
        self._preview_codeview.set_content(textwrap.dedent("""
            def foo(bar):
                if bar is None: # This is a comment
                    print("The answer is", 33)
            
            unclosed_string = "blah, blah
            """).strip())
        self._preview_codeview.grid(row=4, column=0, columnspan=4, sticky=tk.NSEW, pady=(0,5))
        
        (ttk.Label(self, text="NB! Some style elements change only after restarting Thonny")
         .grid(row=5, column=0, columnspan=4, sticky="w", pady=(0, 5)))
            
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.columnconfigure(2, weight=1)
        self.columnconfigure(3, weight=1)
        self.rowconfigure(4, weight=1)
        
    def apply(self):
        # don't do anything, as preview already did the thing
        return
    
    def cancel(self):
        if (getattr(self._family_variable, "modified")
            or getattr(self._size_variable, "modified")
            or getattr(self._ui_theme_variable, "modified")
            or getattr(self._syntax_theme_variable, "modified")):
            get_workbench().set_option("view.ui_theme", self._original_ui_theme)
            get_workbench().set_option("view.syntax_theme", self._original_syntax_theme)
            get_workbench().set_option("view.editor_font_size", self._original_size)
            get_workbench().set_option("view.editor_font_family", self._original_family)
            get_workbench().reload_themes()
            get_workbench().update_fonts()
            
                
    def _update_appearance(self):
        get_workbench().set_option("view.ui_theme", self._ui_theme_variable.get())
        get_workbench().set_option("view.syntax_theme", self._syntax_theme_variable.get())
        get_workbench().set_option("view.editor_font_size", int(self._size_variable.get()))
        get_workbench().set_option("view.editor_font_family", self._family_variable.get())
        get_workbench().reload_themes()
        get_workbench().update_fonts()
    
    def _get_families_to_show(self):
        # In Linux, families may contain duplicates
        return sorted(set(filter(
            lambda name : name[0].isalpha(),          
            tk_font.families()
        )))

def load_plugin() -> None:
    get_workbench().add_configuration_page("Theme & Font", ThemeAndFontConfigurationPage)



from typing import Optional

from thonny import get_workbench
from thonny.misc_utils import running_on_windows
from thonny.ui_utils import scale
from thonny.workbench import UiThemeSettings


def clean(frame_background: str,
          text_background: str,
          normal_detail: str,
          high_detail: str,
          low_detail: str,
          normal_foreground: str,
          high_foreground: str,
          low_foreground: str,
          custom_menubar: Optional[int]=None, # NB! Should be 1 or 0, not True or False (Tk would convert False to "False")
          ) -> UiThemeSettings:
    
    # https://wiki.tcl.tk/37973 (Changing colors)
    # https://github.com/tcltk/tk/blob/master/library/ttk/clamTheme.tcl
    # https://github.com/tcltk/tk/blob/master/generic/ttk/ttkClamTheme.c
    
    
    
    return {
        "." : {
            "configure" : {
                "foreground" : normal_foreground,
                "background" : frame_background,
                "lightcolor" : frame_background,
                "darkcolor"  : frame_background,
                "bordercolor" : frame_background,
                "selectbackground"  : high_detail,
                "selectforeground"  : high_foreground,
            },
            
            "map" : {
                "foreground" : [("disabled", low_foreground),
                                ("active", high_foreground)],
                "background" : [("disabled", frame_background), 
                                ("active", high_detail)],
                "selectbackground" : [("!focus", low_detail)],
                "selectforeground" : [("!focus", normal_foreground)]
            },
        },
        
        "TNotebook" : {
            # https://github.com/tcltk/tk/blob/master/generic/ttk/ttkNotebook.c
            "configure" : {
                "bordercolor" : normal_detail,
                "tabmargins" : [scale(1), 0, 0, 0],     # Margins around tab row
            }
        },
        
        "ButtonNotebook.TNotebook" : {
            "configure" : {
                "bordercolor" : frame_background,
            }
        },
        
        "AutomaticNotebook.TNotebook" : {
            "configure" : {
                "bordercolor" : frame_background,
            }
        },
        
        "TNotebook.Tab" : {
            "configure" : {
                "background" : frame_background,
                "bordercolor" : normal_detail,
            },
            
            "map" : { 
                "background" : [("selected", normal_detail),
                                ("!selected", "!active", frame_background),
                                ("active", "!selected", high_detail)],
                "bordercolor" : [("selected", frame_background),
                                 ("!selected", normal_detail)],
                "lightcolor" : [("selected", normal_detail),
                                ("!selected", frame_background)],
                     
            }
        },
        
        "Treeview" : {
            "configure" : {
                "background" : text_background,
                "borderwidth" : 0,
                "relief" : "flat",
            },
            "map" : {
                "background" : [('selected', 'focus', high_detail),
                                ('selected', '!focus', low_detail)],
                "foreground" : [('selected', 'focus', high_foreground),
                                ('selected', '!focus', normal_foreground)],
            }
        },
        
        "Treeview.Heading" : {
        # https://stackoverflow.com/questions/32051780/how-to-edit-the-style-of-a-heading-in-treeview-python-ttk
            "configure" : {
                "background" : normal_detail,
                "lightcolor" : normal_detail,
                "borderwidth" : 0
            },
            "map" : {
                "background" : [("!active", normal_detail),
                                ("active", normal_detail)]
            }
        },
        
        "TEntry" : {
            "configure" : {
                "fieldbackground" : text_background,
                "lightcolor" : normal_detail,
                "insertcolor" : normal_foreground,
            },
            "map" : {
                "background" : [("readonly", text_background)],
                "bordercolor" : [],
                "lightcolor" : [("focus", high_detail)],
                "darkcolor" : []
            }
        },
        
        "TCombobox" : {
            "configure" : {
                "background" : text_background,
                "fieldbackground" : text_background,
                "selectbackground" : text_background,
                "lightcolor" : text_background,
                "darkcolor" : text_background,
                "bordercolor" : text_background,
                "arrowcolor" : normal_foreground,
                "foreground" : normal_foreground,
                "seleftforeground" : normal_foreground,
                #"padding" : [12,2,12,2],
            },
            "map" : {
                "background" : [("active", text_background)],
                "fieldbackground" : [],
                "selectbackground" : [],
                "selectforeground" : [],
                "foreground" : [],
                "arrowcolor" : []
            }
        },
        
        "TScrollbar" : {
            "configure" : {
                "gripcount" : 0,
                "borderwidth" : 0,
                "relief" : "flat",
                "darkcolor" : normal_detail,
                "lightcolor" : normal_detail,
                "bordercolor" : text_background,
                "troughcolor" : text_background,
                #arrowcolor="white"
            },
            "map" : {
                "background" : [("!disabled", normal_detail), ("disabled", normal_detail)],
                "darkcolor" : [("!disabled", text_background), ("disabled", text_background)],
                "lightcolor" : [("!disabled", text_background), ("disabled", text_background)],
            }
        },
        
        "Vertical.TScrollbar" : {
            # Remove scrollbar buttons/arrows:
            "layout" : [
                ('Vertical.Scrollbar.trough', {'sticky': 'ns', 'children': [
                    ('Vertical.Scrollbar.thumb', {'expand': '1', 'sticky': 'nswe'})
                ]})
            ]
        },
        
        "Horizontal.TScrollbar" : {
            # Remove scrollbar buttons/arrows:
            "layout" : [
                ('Horizontal.Scrollbar.trough', {'sticky': 'we', 'children': [
                    ('Horizontal.Scrollbar.thumb', {'expand': '1', 'sticky': 'nswe'})
                ]})
            ],
            "map" : {
                # Make disabled Hor Scrollbar invisible
                "background" : [("disabled", frame_background), ("!disabled", normal_detail)],
                "troughcolor" : [("disabled", frame_background)],
                "bordercolor" : [("disabled", frame_background)],
                "darkcolor" : [("disabled", frame_background)],
                "lightcolor" : [("disabled", frame_background)],
            }
        },
        
        "TButton" : {
            "configure" : {
                "background" : normal_detail,
                "foreground" : normal_foreground,
            },
            
            "map" : {
                "foreground" : [("disabled", low_foreground),
                                ("alternate", high_foreground)],
                "background" : [("pressed", low_detail),
                                ("active", high_detail)],
                "bordercolor": [("alternate", high_detail)]           
            }
        },
        
        "TCheckbutton" : {
            "configure" : {
                "indicatorforeground" : normal_foreground,
                "indicatorbackground" : text_background,
            },
            "map" : {
                "indicatorforeground" : [("disabled", "alternate", low_foreground),
                                         ("disabled", low_foreground)]
            }
        },
        
        "TRadiobutton" : {
            "configure" : {
                "indicatorforeground" : normal_foreground,
                "indicatorbackground" : text_background,
            },
            "map" : {
                "indicatorforeground" : [("disabled", "alternate", low_foreground),
                                         ("disabled", low_foreground)]
            }
        },
        
        "Toolbutton" : {
            "configure" : {
                "background" : frame_background
            },
            
            "map" : {
                "background": [("disabled", frame_background),
                               ("active", high_detail)]
            }
        },
        
        "TLabel" : {
            "configure"  : {
                "foreground" : normal_foreground
            }
        },
        
        "Url.TLabel" : {
            "configure"  : {
                "foreground" : high_foreground,
            }
        },
        
        "TScale" : {
            "configure" : {
                "background" : high_detail,
                "troughcolor" : normal_detail,
                "lightcolor" : high_detail,
                "darkcolor" : high_detail,
                #"bordercolor" : "red",
                #"sliderlength" : 40,
                #"sliderthickness" : 60,
                "gripcount" : 0,
            },
            "map" : {
                "background" : [],
                "troughcolor" : [],
            }
        },
        
        "TScale.slider" : {
            "configure" : {
                "background" : "red",
                "troughcolor" : "yellow",
                "lightcolor" : "green",
                "darkcolor" : "white",
                #"sliderlength" : 40,
                #"sliderthickness" : 60,
            }
        },
        
        
        "ViewBody.TFrame" : {
            "configure" : {
                "background" : text_background,
            }
        },
        
        "ViewToolbar.TFrame" : {
            "configure" : {
                "background" : normal_detail,
            }
        },
        
        "ViewToolbar.Toolbutton" : {
            "configure" : {
                "background" : normal_detail,
            }
        },
        
        "ViewTab.TLabel" : {
            "configure" : {
                "background" : normal_detail,
                "padding" : [5, 0],
            },
        },
        
        "ViewToolbar.TLabel" : {
            "configure" : {
                "background" : normal_detail,
                "padding" : [scale(4), 0],
            },
        },
        
        "Active.ViewTab.TLabel" : {
            "configure" : {
                "foreground" : high_foreground,
                #"font" : "BoldTkDefaultFont",
                "background" : text_background
            }
        },
        
        "Inactive.ViewTab.TLabel" : {
            "configure" : {
                "foreground" : normal_foreground,
                "font" : "UnderlineTkDefaultFont",
            },
            "map" : {
                "background" : [("hover", high_detail)]
            }
        },
        
        "Text" : {
            "configure" : {
                "background" : text_background,
                "foreground" : normal_foreground
            },
        },
        
        "Gutter" : {
            "configure" : {
                "background" : low_detail,
                "foreground" : low_foreground 
            }
        },
        
        "Listbox" : {
            "configure" : {
                "background" : text_background,
                "foreground" : normal_foreground,
                "selectbackground" : high_detail,
                "selectforeground" : high_foreground,
                "disabledforeground" : low_foreground,
                "highlightbackground" : normal_detail,
                "highlightcolor" : high_detail,
                "highlightthickness" : 1,
            },
        },
        "Menubar" : {
            "configure" : {
                # Regular, system-provided Windows menubar doesn't allow changing colors.
                # custom=True replaces it with a custom-built menubar.
                "custom" : running_on_windows() if custom_menubar is None else custom_menubar, 
                "background" : frame_background,
                "foreground" : normal_foreground,
                "activebackground" : normal_foreground, 
                "activeforeground" : frame_background,
                "relief" : "flat",
            }
        },
        
        "Menu" : {
            "configure" : {
                "background" : normal_detail,
                "foreground" : high_foreground,
                "selectcolor" : normal_foreground,
                "borderwidth" : 0,
                "activebackground" : normal_foreground,
                "activeforeground" : frame_background,
                "activeborderwidth" : 0,
                "relief" : "flat"
            }
        },
        
        "CustomMenubarLabel.TLabel" : {
            "configure" : {
                "padding" : [scale(10),scale(2),0,scale(15)]
            }
        }
    }
    
    
def load_plugin() -> None:
    dark_images = {"tab-close-active" : "tab-close-active-clam-dark"}
    
    get_workbench().add_ui_theme("Clean Dark", "Enhanced Clam", 
        clean(frame_background="#252525",
              text_background="#2d2d2d",
              normal_detail="#3D3D3D",
              high_detail="#6E6E6E",
              low_detail="#404040",
              normal_foreground="#9f9f9f",
              high_foreground="#eeeeee",
              low_foreground="#595959",
        ),
        dark_images
    )

    get_workbench().add_ui_theme("Clean Dark Green", "Enhanced Clam", 
        clean(frame_background="#1D291A",
              text_background="#273627",
              normal_detail="#2D452F",
              high_detail="#3C6E40",
              low_detail="#33402F",
              normal_foreground="#9E9E9E",
              high_foreground="#eeeeee",
              low_foreground="#485C49",
        ),
        dark_images
    )

    get_workbench().add_ui_theme("Clean Dark Blue", "Enhanced Clam", 
        clean(frame_background="#1A1C29",
              text_background="#272936",
              normal_detail="#2D3345",
              high_detail="#3C436E",
              low_detail="#2F3640",
              normal_foreground="#9E9E9E",
              high_foreground="#eeeeee",
              low_foreground="#484A5C",
        ),
        dark_images
    )

    get_workbench().add_ui_theme("Clean Sepia", "Enhanced Clam", 
        clean(frame_background="#E8E7DC",
              text_background="#F7F6F0",
              normal_detail="#DEDCC8",
              high_detail="#eeebe7",
              low_detail="#D4D0B8",
              normal_foreground="#222222",
              high_foreground="#000000",
              low_foreground="#999999",
              custom_menubar=0,
        )
    )



EACCES = 13
EADDRINUSE = 98
EAGAIN = 11
EALREADY = 114
EBADF = 9
ECONNABORTED = 103
ECONNREFUSED = 111
ECONNRESET = 104
EEXIST = 17
EHOSTUNREACH = 113
EINPROGRESS = 115
EINVAL = 22
EIO = 5
EISDIR = 21
ENOBUFS = 105
ENODEV = 19
ENOENT = 2
ENOMEM = 12
ENOTCONN = 107
EOPNOTSUPP = 95
EPERM = 1
ETIMEDOUT = 110
errorcode = None



import ast
import builtins
import re
import token
import tokenize

from thonny.assistance import ErrorHelper, Suggestion, name_similarity,\
    add_error_helper


class SyntaxErrorHelper(ErrorHelper):
    def __init__(self, error_info):
        super().__init__(error_info)
        
        self.tokens = []
        self.token_error = None

        if self.error_info["message"] == "EOL while scanning string literal":
            self.intro_text = ("You haven't properly closed the string on line %s." % self.error_info["lineno"]
                    + "\n(If you want a multi-line string, then surround it with"
                    + " `'''` or `\"\"\"` at both ends.)")
        
        elif self.error_info["message"] == "EOF while scanning triple-quoted string literal":
            # lineno is not useful, as it is at the end of the file and user probably
            # didn't want the string to end there
            self.intro_text = "You haven't properly closed a triple-quoted string"
        
        else:
            self.intro_text = "Python doesn't know how to read your program."
            
            if "^" in self.error_info["message"]:
                self.intro_text += (" Small `^` in the original error message shows where it gave up,"
                        + " but the actual mistake can be before this.") 
            
            if self.error_info["filename"]:
                with open(self.error_info["filename"], mode="rb") as fp:
                    try:
                        for t in tokenize.tokenize(fp.readline):
                            self.tokens.append(t)
                    except tokenize.TokenError as e:
                        self.token_error = e
                
                assert self.tokens[-1].type == token.ENDMARKER
            else:
                self.tokens = None
                
            self.suggestions = [
                self._sug_missing_or_misplaced_colon()
            ]
            
    def _sug_missing_or_misplaced_colon(self):
        i = 0
        title = "Did you forget the colon?"
        relevance = 0
        body = ""
        while i < len(self.tokens) and self.tokens[i].type != token.ENDMARKER:
            t = self.tokens[i]
            if t.string in ["if", "elif", "else", "while", "for", "with",
                            "try", "except", "finally", 
                            "class", "def"]:
                keyword_pos = i
                while (self.tokens[i].type not in [token.NEWLINE, token.ENDMARKER, 
                                     token.COLON, # colon may be OP 
                                     token.RBRACE]
                        and self.tokens[i].string != ":"):
                    
                    old_i = i
                    if self.tokens[i].string in "([{":
                        i = self._skip_braced_part(i)
                        assert i > old_i
                    else:
                        i += 1
                
                if self.tokens[i].string != ":":
                    relevance = 9
                    body = "`%s` header must end with a colon." % t.string
                    break
            
                # Colon was present, but maybe it should have been right
                # after the keyword.
                if (t.string in ["else", "try", "finally"]
                    and self.tokens[keyword_pos+1].string != ":"):
                    title = "Incorrect use of `%s`" % t.string
                    body = "Nothing is allowed between `%s` and colon." % t.string
                    relevance = 9
                    if (self.tokens[keyword_pos+1].type not in (token.NEWLINE, tokenize.COMMENT)
                        and t.string == "else"):
                        body = "If you want to specify a conditon, then use `elif` or nested `if`."
                    break
                
            i += 1
                
        return Suggestion("missing-or-misplaced-colon", title, body, relevance)
    
    def _sug_wrong_increment_op(self):
        pass
    
    def _sug_wrong_decrement_op(self):
        pass
    
    def _sug_wrong_comparison_op(self):
        pass
    
    def _sug_switched_assignment_sides(self):
        pass
    
    def _skip_braced_part(self, token_index):
        assert self.tokens[token_index].string in "([{"
        level = 1
        while token_index < len(self.tokens):
            token_index += 1
            
            if self.tokens[token_index].string in "([{":
                level += 1
            elif self.tokens[token_index].string in ")]}":
                level -= 1
            
            if level <= 0:
                token_index += 1
                return token_index
        
        assert token_index == len(self.tokens)
        return token_index-1
    
    def _find_first_braces_problem(self):
        #closers = {'(':')', '{':'}', '[':']'}
        openers = {')':'(', '}':'{', ']':'['}
        
        brace_stack = []
        for t in self.tokens:
            if t.string in "([{":
                brace_stack.append(token)
            elif t.string in ")]}":
                if not brace_stack:
                    return (t, "`%s` without preceding matching `%s`" % (t.string, openers[t.string]))
                elif brace_stack[-1].string != openers[t.string]:     
                    return (t, "`%s` when last unmatched opener was `%s`" % (t.string, brace_stack[-1].string))
                else:
                    brace_stack.pop()
        
        if brace_stack:
            return (brace_stack[-1], "`%s` was not closed by the end of the program" % brace_stack[-1].string)
        
        return None
        

class NameErrorHelper(ErrorHelper):
    def __init__(self, error_info):
        super().__init__(error_info)
        
        names = re.findall(r"\'.*\'", error_info["message"])
        assert len(names) == 1
        self.name = names[0].strip("'")
        
        self.intro_text = "Python doesn't know what `%s` stands for." % self.name
        self.suggestions = [
            self._sug_bad_spelling(),
            self._sug_missing_quotes(),
            self._sug_missing_import(),
            self._sug_local_from_global(),
            self._sug_not_defined_yet(),
        ]  
    
    def _sug_missing_quotes(self):
        if (self._is_attribute_value() 
            or self._is_call_function()
            or self._is_subscript_value()):
            relevance = 0
        else:
            relevance = 5
            
        return Suggestion(
            "missing-quotes",
            "Did you actually mean string (text)?",
            'If you didn\'t mean a variable but literal text "%s", then surround it with quotes.' % self.name,
            relevance
        )
    
    def _sug_bad_spelling(self):
        
        # Yes, it would be more proper to consult builtins from the backend,
        # but it's easier this way...
        all_names = {name for name in dir(builtins) if not name.startswith("_")}
        all_names |= {"pass", "break", "continue", "return", "yield"}
        
        if self.last_frame.globals is not None:
            all_names |= set(self.last_frame.globals.keys())
        if self.last_frame.locals is not None:
            all_names |= set(self.last_frame.locals.keys())
        
        similar_names = {self.name}
        if all_names:
            relevance = 0
            for name in all_names:
                sim = name_similarity(name, self.name)
                if sim > 4:
                    similar_names.add(name)
                relevance = max(sim, relevance)
        else:
            relevance = 3
        
        if len(similar_names) > 1:
            body = "I found similar names. Are all of them spelled correctly?\n\n"
            for name in sorted(similar_names, key=lambda x: x.lower()):
                # TODO: add location info
                body += "* `%s`\n\n" % name
        else:
            body = (
                "Compare the name with corresponding definition / assignment / documentation."
                + " Don't forget that case of the letters matters!"
            ) 
        
        return Suggestion(
            "bad-spelling-name",
            "Did you misspell it (somewhere)?",
            body,
            relevance
        )
    
    def _sug_missing_import(self):
        likely_importable_functions = {
            "math" : {"ceil", "floor", "sqrt", "sin", "cos", "degrees"},  
            "random" : {"randint"},
            "turtle" : {"left", "right", "forward", "fd", 
                        "goto", "setpos", "Turtle",
                        "penup", "up", "pendown", "down",
                        "color", "pencolor", "fillcolor",
                        "begin_fill", "end_fill", "pensize", "width"},
            "re" : {"search", "match", "findall"},
            "datetime" : {"date", "time", "datetime", "today"},
            "statistics" : {"mean", "median", "median_low", "median_high", "mode", 
                            "pstdev", "pvariance", "stdev", "variance"},
            "os" : {"listdir"},
            "time" : {"time", "sleep"},
        }
        
        body = None
         
        if self._is_call_function():
            relevance = 5
            for mod in likely_importable_functions:
                if self.name in likely_importable_functions[mod]:
                    relevance += 3
                    body = ("If you meant `%s` from module `%s`, then add\n\n`from %s import %s`\n\nto the beginning of your script."
                                % (self.name, mod, mod, self.name))
                    break
                
        elif self._is_attribute_value():
            relevance = 5
            body = ("If you meant module `%s`, then add `import %s` to the beginning of your script"
                        % (self.name, self.name))
            
            if self.name in likely_importable_functions:
                relevance += 3
                
                
        elif self._is_subscript_value() and self.name != "argv":
            relevance = 0
        elif self.name == "pi":
            body = "If you meant the constant π, then add `from math import pi` to the beginning of your script."
            relevance = 8
        elif self.name == "argv":
            body = "If you meant the list with program arguments, then add `from sys import argv` to the beginning of your script."
            relevance = 8
        else:
            relevance = 3
            
        
        if body is None:
            body = "Some functions/variables need to be imported before they can be used."
            
        return Suggestion("missing-import",
                          "Did you forget to import it?",
                          body,
                          relevance)
    
    def _sug_local_from_global(self):
        relevance = 0
        body = None
        
        if (self.last_frame.code_name == "<module>"
            and self.last_frame_module_ast is not None):
            function_names = set()
            for node in ast.walk(self.last_frame_module_ast):
                if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                    if self.name in map(lambda x: x.arg, node.args.args):
                        function_names.add(node.name)
                    # TODO: varargs, kw, ...
                    declared_global = False
                    for localnode in ast.walk(node):
                        #print(node.name, localnode)
                        if (isinstance(localnode, ast.Name)
                            and localnode.id == self.name
                            and isinstance(localnode.ctx, ast.Store)
                            ):
                            function_names.add(node.name)
                        elif (isinstance(localnode, ast.Global)
                              and self.name in localnode.names):
                            declared_global = True
                    
                    if node.name in function_names and declared_global:
                        function_names.remove(node.name)
            
            if function_names:
                relevance = 9
                body = (
                    ("Name `%s` defined in `%s` is not accessible in the global/module level."
                     % (self.name, " and ".join(function_names)))
                    + "\n\nIf you need that data at the global level, then consider changing the function so that it `return`-s the value.")
            
        return Suggestion(
            "local-from-global",
            "Are you trying to acces a local variable outside of the function?",
            body,
            relevance)
    
    def _sug_not_defined_yet(self):
        return Suggestion(
            "not-defined-yet",
            "Has Python executed the definition?",
            ("Don't forget that name becomes defined when corresponding definition ('=', 'def' or 'import') gets executed."
            + " If the definition comes later in code or is inside an if-statement, Python may not have executed it (yet)."
            + "\n\n"
            + "Make sure Python arrives to the definition before it arrives to this line. When in doubt, use the debugger."),
            2)
    
    def _sug_maybe_attribute(self):
        "TODO:"
    
    def _sug_synonym(self):
        "TODO:"
    
    def _is_call_function(self):
        return self.name + "(" in (self.error_info["line"]
                                   .replace(" ", "")
                                   .replace("\n", "")
                                   .replace("\r", ""))
                                   
    def _is_subscript_value(self):
        return self.name + "[" in (self.error_info["line"]
                                   .replace(" ", "")
                                   .replace("\n", "")
                                   .replace("\r", ""))
                                   
    def _is_attribute_value(self):
        return self.name + "." in (self.error_info["line"]
                                   .replace(" ", "")
                                   .replace("\n", "")
                                   .replace("\r", ""))

class AttributeErrorHelper(ErrorHelper):
    def __init__(self, error_info):
        super().__init__(error_info)
        
        names = re.findall(r"\'.*?\'", error_info["message"])
        assert len(names) == 2
        self.type_name = names[0].strip("'")
        self.att_name = names[1].strip("'")
        
        self.intro_text = (
            "Your program tries to "
            + ("call method " 
               if self._is_call_function() 
               else "access attribute ") 
            + "`%s` of " % self.att_name
            + _get_phrase_for_object(self.type_name)
            + ", but this type doesn't have such "
            + ("method." 
               if self._is_call_function() 
               else "attribute.")
        )
        
        self.suggestions = [
            self._sug_wrong_attribute_instead_of_len(),
            self._sug_bad_spelling(),
            self._sug_bad_type()
        ] 
    
    def _sug_wrong_attribute_instead_of_len(self):
        
        if self.type_name == "str":
            goal = "length"
        elif self.type_name == "bytes":
            goal = "number of bytes"
        elif self.type_name == "list":
            goal = "number of elements"
        elif self.type_name == "tuple":
            goal = "number of elements"
        elif self.type_name == "set":
            goal = "number of elements"
        elif self.type_name == "dict":
            goal = "number of entries"
        else:
            return
         
        return Suggestion(
            "wrong-attribute-instead-of-len",
            "Did you mean to ask the %s?" % goal,
            "This can be done with function `len`, eg:\n\n`len(%s)`" 
                % _get_sample_for_type(self.type_name),
            (9
             if self.att_name.lower() in ("len", "length", "size")
             else 0) 
        )
    
    def _sug_bad_spelling(self):
        # TODO: compare with attributes of known types
        return Suggestion(
            "bad-spelling-attribute",
            "Did you misspell the name?",
            "Don't forget that case of the letters matters too!",
            3 
        )
    
    def _sug_bad_type(self):
        if self._is_call_function():
            action = "call this function on"
        else:
            action = "ask this attribute from"
             
        return Suggestion(
            "wrong-type-attribute",
            "Did you expect another type?",
            "If you didn't mean %s %s, " 
                % (action, _get_phrase_for_object(self.type_name))
                + "then step through your program to see "
                + "why this type appears here.",
            3 
        )
    
        
    def _is_call_function(self):
        return "." + self.att_name + "(" in (self.error_info["line"]
                                   .replace(" ", "")
                                   .replace("\n", "")
                                   .replace("\r", ""))
        

class TypeErrorHelper(ErrorHelper):
    def __init__(self, error_info):
        super().__init__(error_info)
        
        self.intro_text = (
            "Python was asked to do an operation with an object which "
            + "doesn't support it."
        )
        
        self.suggestions = [
            Suggestion(
                "step-to-find-type-error",
                "Did you expect another type?",
                "Step through your program to see why this type appears here.",
                3
            ),
            Suggestion(
                "look-documentation-type-error",
                "Maybe you forgot some details about this operation?",
                "Look up the documentation or perform a web search with the error message.",
                2
            )
        ]
        
        
        # overwrite / add for special cases
        # something + str or str + something
        for r, string_first in [
            (r"unsupported operand type\(s\) for \+: '(.+?)' and 'str'", False),
            (r"^Can't convert '(.+?)' object to str implicitly$", True), # Python 3.5
            (r"^must be str, not (.+)$", True), # Python 3.6
            (r'^can only concatenate str (not "(.+?)") to str$', True), # Python 3.7
            ]:
            m = re.match(r, error_info["message"], re.I)  # @UndefinedVariable
            if m is not None:
                self._bad_string_concatenation(m.group(1), string_first)
                return
    
    def _bad_string_concatenation(self, other_type_name, string_first):
        self.intro_text = (
            "Your program is trying to put together "
            + ("a string and %s." if string_first else "%s and a string.")
                % _get_phrase_for_object(other_type_name)
        )
        
        self.suggestions.append(Suggestion(
            "convert-other-operand-to-string",
            "Did you mean to treat both sides as text and produce a string?",
            "In this case you should apply function `str` to the %s " 
                % _get_phrase_for_object(other_type_name, False)
            + "in order to convert it to string first, eg:\n\n"
            + ("`'abc' + str(%s)`" if string_first else "`str(%s) + 'abc'`")
                % _get_sample_for_type(other_type_name),
            
            8
        ))
        
        if other_type_name in ("float", "int"):
            self.suggestions.append(Suggestion(
                "convert-other-operand-to-number",
                "Did you mean to treat both sides as numbers and produce a sum?",
                "In this case you should first convert the string to a number "
                + "using either function `float` or `int`, eg:\n\n"
                + ("`float('3.14') + 22`" if string_first else "`22 + float('3.14')`"),
                7
            ))
        

def _get_phrase_for_object(type_name, with_article=True):
    friendly_names = {
        "str" : "a string",
        "int" : "an integer",
        "float" : "a float",
        "list" : "a list",
        "tuple" : "a tuple",
        "dict" : "a dictionary",
        "set" : "a set",
        "bool" : "a boolean"
    }
    result = friendly_names.get(type_name,
                              "an object of type '%s'" % type_name)
    
    if with_article:
        return result
    else:
        _, rest = result.split(" ", maxsplit=1)
        return rest

def _get_sample_for_type(type_name):
    if type_name == "int":
        return "42"
    elif type_name == "float":
        return "3.14"
    elif type_name == "str":
        return "'abc'"
    elif type_name == "bytes":
        return "b'abc'"
    elif type_name == "list":
        return "[1, 2, 3]"
    elif type_name == "tuple":
        return "(1, 2, 3)"
    elif type_name == "set":
        return "{1, 2, 3}"
    elif type_name == "dict":
        return "{1 : 'one', 2 : 'two'}"
    else:
        return "..."
    


def load_plugin():
    for name in globals():
        if name.endswith("ErrorHelper") and not name.startswith("_"):
            type_name = name[:-len("Helper")]
            add_error_helper(type_name, globals()[name])
            


# -*- coding: utf-8 -*-

import ast
import io
import logging
import sys
import token
import tokenize

import _ast


def extract_text_range(source, text_range):
    if isinstance(source, bytes):
        # TODO: may be wrong encoding
        source = source.decode("utf-8")
        
    lines = source.splitlines(True)
    # get relevant lines
    lines = lines[text_range.lineno-1:text_range.end_lineno]

    # trim last and first lines
    lines[-1] = lines[-1][:text_range.end_col_offset]
    lines[0] = lines[0][text_range.col_offset:]
    return "".join(lines)



def find_expression(node, text_range):
    if (hasattr(node, "lineno")
        and node.lineno == text_range.lineno 
        and node.col_offset == text_range.col_offset
        and node.end_lineno == text_range.end_lineno 
        and node.end_col_offset == text_range.end_col_offset
        # expression and Expr statement can have same range
        and isinstance(node, _ast.expr)):
        return node
    else:
        for child in ast.iter_child_nodes(node):
            result = find_expression(child, text_range)
            if result is not None:
                return result
        return None


def contains_node(parent_node, child_node):
    for child in ast.iter_child_nodes(parent_node):
        if child == child_node or contains_node(child, child_node):
            return True

    return False

def has_parent_with_class(target_node, parent_class, tree):
    for node in ast.walk(tree):
        if isinstance(node, parent_class) and contains_node(node, target_node):
            return True

    return False


def parse_source(source: bytes, filename='<unknown>', mode="exec"):
    root = ast.parse(source, filename, mode)
    mark_text_ranges(root, source)
    return root


def get_last_child(node, skip_incorrect=True):
    
    def ok_node(node):
        if node is None:
            return None
        
        if skip_incorrect and getattr(node, "incorrect_range", False):
            return None
        
        return node
    
    def last_ok(nodes):
        for i in range(len(nodes)-1, -1, -1):
            if ok_node(nodes[i]):
                node = nodes[i]
                if isinstance(node, ast.Starred):
                    if ok_node(node.value):
                        return node.value
                    else:
                        return None
                else:
                    return nodes[i]
            
        return None
    
    if isinstance(node, ast.Call):
        # TODO: take care of Python 3.5 updates (Starred etc.)
        if hasattr(node, "kwargs") and ok_node(node.kwargs):
            return node.kwargs
        elif hasattr(node, "starargs") and ok_node(node.starargs):
            return node.starargs
        elif last_ok(node.keywords):
            return last_ok(node.keywords)
        elif last_ok(node.args):
            return last_ok(node.args)
        else:
            return ok_node(node.func)

    elif isinstance(node, ast.BoolOp):
        return last_ok(node.values)

    elif isinstance(node, ast.BinOp):
        if ok_node(node.right):
            return node.right
        else:
            return ok_node(node.left)

    elif isinstance(node, ast.Compare):
        return last_ok(node.comparators)

    elif isinstance(node, ast.UnaryOp):
        return ok_node(node.operand)

    elif isinstance(node, (ast.Tuple, ast.List, ast.Set)):
        return last_ok(node.elts)

    elif isinstance(node, ast.Dict):
        # TODO: actually should pairwise check last value, then last key, etc.
        return last_ok(node.values)

    elif isinstance(node, (ast.Return, ast.Assign, ast.AugAssign, ast.Yield, ast.YieldFrom)):
        return ok_node(node.value)

    elif isinstance(node, ast.Delete):
        return last_ok(node.targets)

    elif isinstance(node, ast.Expr):
        return ok_node(node.value)

    elif isinstance(node, ast.Assert):
        if ok_node(node.msg):
            return node.msg
        else:
            return ok_node(node.test)

    elif isinstance(node, ast.Subscript):
        if hasattr(node.slice, "value") and ok_node(node.slice.value):
            return node.slice.value
        else:
            assert (hasattr(node.slice, "lower")
                    and hasattr(node.slice, "upper")
                    and hasattr(node.slice, "step"))

            if ok_node(node.slice.step):
                return node.slice.step
            elif ok_node(node.slice.upper):
                return node.slice.upper
            else:
                return ok_node(node.slice.lower)


    elif isinstance(node, (ast.For, ast.While, ast.If, ast.With)):
        return True # There is last child, but I don't know which it will be


    # TODO: pick more cases from here:
    """
    (isinstance(node, (ast.IfExp, ast.ListComp, ast.SetComp, ast.DictComp, ast.GeneratorExp))
            or isinstance(node, ast.Raise) and (node.exc is not None or node.cause is not None)
            # or isinstance(node, ast.FunctionDef, ast.Lambda) and len(node.args.defaults) > 0
                and (node.dest is not None or len(node.values) > 0))

            #"TODO: Import ja ImportFrom"
            # TODO: what about ClassDef ???
    """

    return None





def mark_text_ranges(node, source: bytes):
    """
    Node is an AST, source is corresponding source as string.
    Function adds recursively attributes end_lineno and end_col_offset to each node
    which has attributes lineno and col_offset.
    """


    def _extract_tokens(tokens, lineno, col_offset, end_lineno, end_col_offset):
        return list(filter((lambda tok: tok.start[0] >= lineno
                                   and (tok.start[1] >= col_offset or tok.start[0] > lineno)
                                   and tok.end[0] <= end_lineno
                                   and (tok.end[1] <= end_col_offset or tok.end[0] < end_lineno)
                                   and tok.string != ''),
                           tokens))



    def _mark_text_ranges_rec(node, tokens, prelim_end_lineno, prelim_end_col_offset):
        """
        Returns the earliest starting position found in given tree,
        this is convenient for internal handling of the siblings
        """

        # set end markers to this node
        if "lineno" in node._attributes and "col_offset" in node._attributes:
            tokens = _extract_tokens(tokens, node.lineno, node.col_offset, prelim_end_lineno, prelim_end_col_offset)
            try:
                tokens = _mark_end_and_return_child_tokens(node, tokens, prelim_end_lineno, prelim_end_col_offset)
            except Exception:
                logging.getLogger("thonny").warning("Problem with marking %s", node)
                # fallback to incorrect marking instead of exception
                node.incorrect_range = True
                node.end_lineno = node.lineno
                node.end_col_offset = node.col_offset + 1


        # mark its children, starting from last one
        # NB! need to sort children because eg. in dict literal all keys come first and then all values
        children = list(_get_ordered_child_nodes(node))
        for child in reversed(children):
            (prelim_end_lineno, prelim_end_col_offset) = \
                _mark_text_ranges_rec(child, tokens, prelim_end_lineno, prelim_end_col_offset)

        if "lineno" in node._attributes and "col_offset" in node._attributes:
            # new "front" is beginning of this node
            prelim_end_lineno = node.lineno
            prelim_end_col_offset = node.col_offset

        return (prelim_end_lineno, prelim_end_col_offset)


    def _strip_trailing_junk_from_expressions(tokens):
        while (tokens[-1].type not in (token.RBRACE, token.RPAR, token.RSQB,
                                      token.NAME, token.NUMBER, token.STRING,
                                      token.ELLIPSIS)
                    and tokens[-1].string != "..." # See https://bugs.python.org/issue31394
                    and tokens[-1].string not in ")}]"
                    or tokens[-1].string in ['and', 'as', 'assert', 'class', 'def', 'del',
                                              'elif', 'else', 'except', 'finally',
                                              'for', 'from', 'global', 'if', 'import', 'in',
                                              'is', 'lambda', 'not', 'or', 'try',
                                              'while', 'with', 'yield']):
            del tokens[-1]

    def _strip_trailing_extra_closers(tokens, remove_naked_comma):
        level = 0
        for i in range(len(tokens)):
            if tokens[i].string in "({[":
                level += 1
            elif tokens[i].string in ")}]":
                level -= 1

            if level == 0 and tokens[i].string == "," and remove_naked_comma:
                tokens[:] = tokens[0:i]
                return

            if level < 0:
                tokens[:] = tokens[0:i]
                return

    def _strip_unclosed_brackets(tokens):
        level = 0
        for i in range(len(tokens)-1, -1, -1):
            if tokens[i].string in "({[":
                level -= 1
            elif tokens[i].string in ")}]":
                level += 1

            if level < 0:
                tokens[:] = tokens[0:i]
                level = 0  # keep going, there may be more unclosed brackets

    def _mark_end_and_return_child_tokens(node, tokens, prelim_end_lineno, prelim_end_col_offset):
        """
        # shortcut
        node.end_lineno = prelim_end_lineno
        node.end_col_offset = prelim_end_col_offset
        return tokens
        """
        # prelim_end_lineno and prelim_end_col_offset are the start of
        # next positioned node or end of source, ie. the suffix of given
        # range may contain keywords, commas and other stuff not belonging to current node

        # Function returns the list of tokens which cover all its children


        if isinstance(node, _ast.stmt):
            # remove empty trailing lines
            while (tokens[-1].type in (tokenize.NL, tokenize.COMMENT, token.NEWLINE, token.INDENT)
                   or tokens[-1].string in (":", "else", "elif", "finally", "except")):
                del tokens[-1]

        else:
            _strip_trailing_extra_closers(tokens, not isinstance(node, (ast.Tuple, ast.Lambda)))
            _strip_trailing_junk_from_expressions(tokens)
            _strip_unclosed_brackets(tokens)

        # set the end markers of this node
        node.end_lineno = tokens[-1].end[0]
        node.end_col_offset = tokens[-1].end[1]

        # Peel off some trailing tokens which can't be part any
        # positioned child node.
        # TODO: maybe cleaning from parent side is better than
        # _strip_trailing_junk_from_expressions

        # Remove trailing empty parens from no-arg call
        if (isinstance(node, ast.Call)
            and _tokens_text(tokens[-2:]) == "()"):
            del tokens[-2:]

        # Remove trailing full slice
        elif isinstance(node, ast.Subscript):
            if  _tokens_text(tokens[-3:]) == "[:]":
                del tokens[-3:]

            elif _tokens_text(tokens[-4:]) == "[::]":
                del tokens[-4:]

        # Attribute name would confuse the "value" of Attribute
        elif isinstance(node, ast.Attribute):
            assert tokens[-1].type == token.NAME
            del tokens[-1]
            _strip_trailing_junk_from_expressions(tokens)

        return tokens

    all_tokens = list(tokenize.tokenize(io.BytesIO(source).readline))
    source_lines = source.splitlines(True)
    fix_ast_problems(node, source_lines, all_tokens)
    prelim_end_lineno = len(source_lines)
    prelim_end_col_offset = len(source_lines[len(source_lines)-1])
    _mark_text_ranges_rec(node, all_tokens, prelim_end_lineno, prelim_end_col_offset)


def fix_ast_problems(tree, source_lines, tokens):
    # Problem 1:
    # Python parser gives col_offset as offset to its internal UTF-8 byte array
    # I need offsets to chars
    
    # TODO: what if source_lines are in a different encoding???
    utf8_byte_lines = source_lines

    # Problem 2:
    # triple-quoted strings have just plain wrong positions: http://bugs.python.org/issue18370
    # Fortunately lexer gives them correct positions
    string_tokens = list(filter(lambda tok: tok.type == token.STRING, tokens))

    # Problem 3:
    # Binary operations have wrong positions: http://bugs.python.org/issue18374

    # Problem 4:
    # Function calls have wrong positions in Python 3.4: http://bugs.python.org/issue21295
    # TODO: Python 3.4 is not supported anymore
    # similar problem is with Attributes and Subscripts
    
    # Problem 5: FormattedValue nodes have location of corresponding 
    # JoinedStr, but I'd prefer being more precise 

    JoinedStr = getattr(ast, "JoinedStr", type(None))
    FormattedValue = getattr(ast, "FormattedValue", type(None))

    def fix_node(node, inside_joined_str=False):
        
        # first fix the children
        for child in _get_ordered_child_nodes(node):
            fix_node(child, isinstance(node, JoinedStr) or inside_joined_str)

        # now the node itself
        if isinstance(node, (ast.Str, JoinedStr)) and not inside_joined_str:
            # fix triple-quote problem
            # get position from tokens.
            
            # (Child Str positions are wrong in 3.7, but I don't know how to fix them) 
            # Don't recurse inside JoinedStr as it may have several child Str nodes 
            # but only one string token. 
            # TODO: implicit concatenation messes up token-node correspondence?
            tok = string_tokens.pop(0)
            node.lineno, node.col_offset = tok.start
        
        elif isinstance(node, FormattedValue):
            # Node has wrong position in 3.7, probably taken from string token.
            # Use the position of value instead
            node.lineno = node.value.lineno
            node.col_offset = node.value.col_offset

        elif (isinstance(node, (ast.Expr, ast.Attribute))
            and isinstance(node.value, ast.Str)):
            # they share the wrong offset of their triple-quoted child
            # get position from already fixed child
            # TODO: try whether this works when child is in parentheses
            node.lineno = node.value.lineno
            node.col_offset = node.value.col_offset

        elif (isinstance(node, ast.BinOp)
            and compare_node_positions(node, node.left) > 0):
            # fix binop problem
            # get position from an already fixed child
            node.lineno = node.left.lineno
            node.col_offset = node.left.col_offset
            
            # Note that this doesn't fix
            # (3)+3

        elif (isinstance(node, ast.Call)
            and compare_node_positions(node, node.func) > 0):
            # Python 3.4 call problem
            # get position from an already fixed child
            node.lineno = node.func.lineno
            node.col_offset = node.func.col_offset
            
            # Note that there remains problem with
            # (f)()
            # but this can't be fixed without knowing child end position

        elif (isinstance(node, ast.Attribute)
            and compare_node_positions(node, node.value) > 0):
            # Python 3.4 attribute problem ...
            node.lineno = node.value.lineno
            node.col_offset = node.value.col_offset

        elif (isinstance(node, ast.Subscript)
            and compare_node_positions(node, node.value) > 0):
            # Python 3.4 Subscript problem ...
            node.lineno = node.value.lineno
            node.col_offset = node.value.col_offset

        else:
            # Let's hope this node has correct lineno, and byte-based col_offset
            # Now compute char-based col_offset
            if hasattr(node, "lineno"):
                byte_line = utf8_byte_lines[node.lineno-1]
                char_col_offset = len(byte_line[:node.col_offset].decode("UTF-8"))
                node.col_offset = char_col_offset


    fix_node(tree)

def compare_node_positions(n1, n2):
    if n1.lineno > n2.lineno:
        return 1
    elif n1.lineno < n2.lineno:
        return -1
    elif n1.col_offset > n2.col_offset:
        return 1
    elif n1.col_offset < n2.col_offset:
        return -1
    else:
        return 0


def pretty(node, key="/", level=0):
    """Used for testing and new test generation via AstView.
    Don't change the format without updating tests"""
    if isinstance(node, ast.AST):
        fields = [(key, val) for key, val in ast.iter_fields(node)]
        value_label = node.__class__.__name__
        if isinstance(node, ast.Call):
            # Try to make 3.4 AST-s more similar to 3.5
            if sys.version_info[:2] == (3,4):
                if ("kwargs", None) in fields:
                    fields.remove(("kwargs", None))
                if ("starargs", None) in fields:
                    fields.remove(("starargs", None))

            # TODO: translate also non-None kwargs and starargs

    elif isinstance(node, list):
        fields = list(enumerate(node))
        if len(node) == 0:
            value_label = "[]"
        else:
            value_label = "[...]"
    else:
        fields = []
        value_label = repr(node)

    item_text = level * '    ' + str(key) + "=" + value_label

    if hasattr(node, "lineno"):
        item_text += " @ " + str(getattr(node, "lineno"))
        if hasattr(node, "col_offset"):
            item_text += "." + str(getattr(node, "col_offset"))

        if hasattr(node, "end_lineno"):
            item_text += "  -  " + str(getattr(node, "end_lineno"))
            if hasattr(node, "end_col_offset"):
                item_text += "." + str(getattr(node, "end_col_offset"))

    lines = [item_text] + [pretty(field_value, field_key, level+1)
                           for field_key, field_value in fields]

    return "\n".join(lines)


def _get_ordered_child_nodes(node):
    if isinstance(node, ast.Dict):
        children = []
        for i in range(len(node.keys)):
            children.append(node.keys[i])
            children.append(node.values[i])
        return children
    elif isinstance(node, ast.Call):
        children = [node.func] + node.args

        for kw in node.keywords:
            children.append(kw.value)

        # TODO: take care of Python 3.5 updates (eg. args=[Starred] and keywords)
        if hasattr(node, "starargs") and node.starargs is not None:
            children.append(node.starargs)
        if hasattr(node, "kwargs") and node.kwargs is not None:
            children.append(node.kwargs)

        children.sort(key=lambda x: (x.lineno, x.col_offset))
        return children

    # arguments and their defaults are detached in the AST
    elif isinstance(node, ast.arguments):
        children = node.args + node.kwonlyargs + node.kw_defaults + node.defaults

        if node.vararg is not None:
            children.append(node.vararg)
        if node.kwarg is not None:
            children.append(node.kwarg)

        children.sort(key=lambda x: (x.lineno, x.col_offset))
        return children

    else:
        return ast.iter_child_nodes(node)

def _tokens_text(tokens):
    return "".join([t.string for t in tokens])



from thonny.assistance import ProgramAnalyzer, add_program_analyzer
from thonny import get_runner, rst_utils
from thonny.running import CPythonProxy
import logging
import os
from thonny.common import is_same_path


known_stdlib_modules = {
    # Compiled from https://docs.python.org/3.7/py-modindex.html
    '__future__', '__main__', '_dummy_thread', '_thread', 'abc', 'aifc',
    'argparse', 'array', 'ast', 'asynchat', 'asyncio', 'asyncore', 'atexit',
    'audioop', 'base64', 'bdb', 'binascii', 'binhex', 'bisect', 'builtins', 'bz2',
    'calendar', 'cgi', 'cgitb', 'chunk', 'cmath', 'cmd', 'code', 'codecs',
    'codeop', 'collections', 'colorsys', 'compileall', 'concurrent',
    'configparser', 'contextlib', 'contextvars', 'copy', 'copyreg', 'cProfile',
    'crypt', 'csv', 'ctypes', 'curses', 'dataclasses', 'datetime', 'dbm',
    'decimal', 'difflib', 'dis', 'distutils', 'doctest', 'dummy_threading',
    'email', 'encodings', 'ensurepip', 'enum', 'errno', 'faulthandler', 'fcntl',
    'filecmp', 'fileinput', 'fnmatch', 'formatter', 'fractions', 'ftplib',
    'functools', 'gc', 'getopt', 'getpass', 'gettext', 'glob', 'grp', 'gzip',
    'hashlib', 'heapq', 'hmac', 'html', 'http', 'imaplib', 'imghdr', 'imp',
    'importlib', 'inspect', 'io', 'ipaddress', 'itertools', 'json', 'keyword',
    'lib2to3', 'linecache', 'locale', 'logging', 'lzma', 'macpath', 'mailbox',
    'mailcap', 'marshal', 'math', 'mimetypes', 'mmap', 'modulefinder', 'msilib',
    'msvcrt', 'multiprocessing', 'netrc', 'nis', 'nntplib', 'numbers', 'operator',
    'optparse', 'os', 'ossaudiodev', 'parser', 'pathlib', 'pdb', 'pickle',
    'pickletools', 'pipes', 'pkgutil', 'platform', 'plistlib', 'poplib', 'posix',
    'pprint', 'profile', 'pstats', 'pty', 'pwd', 'py_compile', 'pyclbr', 'pydoc',
    'queue', 'quopri', 'random', 're', 'readline', 'reprlib', 'resource',
    'rlcompleter', 'runpy', 'sched', 'secrets', 'select', 'selectors', 'shelve',
    'shlex', 'shutil', 'signal', 'site', 'smtpd', 'smtplib', 'sndhdr', 'socket',
    'socketserver', 'spwd', 'sqlite3', 'ssl', 'stat', 'statistics', 'string',
    'stringprep', 'struct', 'subprocess', 'sunau', 'symbol', 'symtable', 'sys',
    'sysconfig', 'syslog', 'tabnanny', 'tarfile', 'telnetlib', 'tempfile',
    'termios', 'test', 'textwrap', 'threading', 'time', 'timeit', 'tkinter',
    'token', 'tokenize', 'trace', 'traceback', 'tracemalloc', 'tty', 'turtle',
    'turtledemo', 'types', 'typing', 'unicodedata', 'unittest', 'urllib', 'uu',
    'uuid', 'venv', 'warnings', 'wave', 'weakref', 'webbrowser', 'winreg',
    'winsound', 'wsgiref', 'xdrlib', 'xml', 'xmlrpc', 'zipapp', 'zipfile',
    'zipimport', 'zlib'
}

class ProgramNamingAnalyzer(ProgramAnalyzer):
    def start_analysis(self, main_file_path, imported_file_paths):
        self.completion_handler(self, list(self._get_warnings(main_file_path)))
    
    def _get_warnings(self, main_file_path):
        # TODO: current dir may be different 
        main_file_dir = os.path.dirname(main_file_path)
        if not os.path.isdir(main_file_dir):
            return []
        
        library_modules = known_stdlib_modules | self._get_3rd_party_modules()
        
        for item in os.listdir(main_file_dir):
            full_path = os.path.join(main_file_dir, item)
            if (item.endswith(".py")
                and item[:-3] in library_modules):
                
                if is_same_path(full_path, main_file_path):
                    prelude = "Your program file is named '%s'." % item
                    rename_hint = " (*Run → Rename…* )"
                else:
                    prelude = (
                        "Your working directory `%s <%s>`__ contains a file named '%s'.\n\n"
                        % (rst_utils.escape(main_file_dir),
                               rst_utils.escape(main_file_dir),
                               item)
                    )
                    rename_hint = ""
                
                yield {
                    "filename" : full_path,
                    "lineno" : 0,
                    "symbol" : "file-shadows-library-module",
                    "msg" : "Bad file name",
                    "explanation_rst" : 
                        prelude
                        + "\n\n"                            
                        + "When you try to import library module ``%s``, your file will be imported instead.\n\n"
                            % item[:-3]
                        + "Rename your '%s'%s to make the library module visible again."
                            % (item, rename_hint),
                    "group" : "warnings",
                    "relevance" : 5
                }
    
    def _get_3rd_party_modules(self):
        proxy = get_runner().get_backend_proxy()
        if not isinstance(proxy, CPythonProxy):
            return []
        
        try:
            sys_path = proxy.get_sys_path()
        except Exception:
            logging.exception("Can't get sys path from proxy")
            return []
        
        module_names = set() 
        for item in sys_path:
            if (os.path.isdir(item)
                and ("site-packages" in item or "dist-packages" in item)):
                module_names.update(self._get_module_names(item))
                for name in os.listdir(item):
                    if "-" not in name:
                        module_names.add(name.replace(".py",""))
        
        return module_names
    
    def _get_module_names(self, dir_path):
        result = set()
        for name in os.listdir(dir_path):
            if "-" not in name:
                result.add(name.replace(".py",""))
        return result

def load_plugin():
    add_program_analyzer(ProgramNamingAnalyzer)



# -*- coding: utf-8 -*-
 
"""
Classes used both by front-end and back-end
"""
import os.path
import tokenize
from collections import namedtuple
from typing import List, Optional  # @UnusedImport

MESSAGE_MARKER = "\x02"

ValueInfo = namedtuple('ValueInfo', ['id', 'repr'])
FrameInfo = namedtuple("FrameInfo", ["id", "filename", "module_name", "code_name",
                                     "source", "lineno", "firstlineno", "in_library",
                                     "locals", "globals", "freevars", 
                                     "event", "focus", "node_tags",
                                     "current_statement", 
                                     "current_root_expression", "current_evaluations"]) 

TextRange = namedtuple("TextRange", ["lineno", "col_offset", "end_lineno", "end_col_offset"])
    
class Record:
    def __init__(self, **kw):
        self.__dict__.update(kw)
    
    def update(self, e, **kw):
        self.__dict__.update(e, **kw)
    
    def setdefault(self, **kw):
        "updates those fields that are not yet present (similar to dict.setdefault)"
        for key in kw:
            if not hasattr(self, key):
                setattr(self, key, kw[key])
    
    def get(self, key, default=None):
        return self.__dict__.get(key, default)
    
    def __getitem__(self, key):
        return self.__dict__[key]
    
    def __delitem__(self, key):
        self.__dict__.__delitem__(key)
    
    def __setitem__(self, key, value):
        self.__dict__[key] = value
    
    def __contains__(self, key):
        return key in self.__dict__
    
    def __repr__(self):
        keys = self.__dict__.keys()
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in keys)
        return "{}({})".format(self.__class__.__name__, ", ".join(items))
    
    def __str__(self):
        keys = sorted(self.__dict__.keys())
        items = ("{}={!r}".format(k, str(self.__dict__[k])) for k in keys)
        return "{}({})".format(self.__class__.__name__, ", ".join(items))
    
    def __eq__(self, other):
        # pylint: disable=unidiomatic-typecheck
        
        if type(self) != type(other):
            return False
        
        if len(self.__dict__) != len(other.__dict__):
            return False 
        
        for key in self.__dict__:
            if not hasattr(other, key):
                return False
            self_value = getattr(self, key)
            other_value = getattr(other, key)
            
            if type(self_value) != type(other_value) or self_value != other_value:
                return False
        
        return True

    def __ne__(self, other):
        return not self.__eq__(other)
        
    def __hash__(self):
        return hash(repr(self))


def range_contains_smaller(one: TextRange, other: TextRange) -> bool:
    this_start = (one.lineno, one.col_offset)
    this_end = (one.end_lineno, one.end_col_offset)
    other_start = (other.lineno, other.col_offset)
    other_end = (other.end_lineno, other.end_col_offset)
    
    return (this_start < other_start and this_end > other_end
            or this_start == other_start and this_end > other_end
            or this_start < other_start and this_end == other_end)

def range_contains_smaller_or_equal(one: TextRange, other: TextRange) -> bool:
    return range_contains_smaller(one, other) or one == other
    
class InputSubmission(Record):
    """For sending data to backend's stdin"""
    def __init__(self, data: str, **kw) -> None:
        super().__init__(**kw)
        self.data = data

class CommandToBackend(Record):
    """Command meant for the back-end"""
    
    def __init__(self, name: str, **kw) -> None:
        super().__init__(**kw)
        self.name = name

class ToplevelCommand(CommandToBackend):
    def __init__(self, name: str, argv: List[str]=[], **kw)->None:
        super().__init__(name, **kw)

class DebuggerCommand(CommandToBackend):
    pass

class InlineCommand(CommandToBackend):
    """
    Can be used both during debugging and in waiting_toplevel_command state
    (eg. for sending variable and heap info requests)
    """
    pass


class MessageFromBackend(Record):
    def __init__(self, **kw) -> None:
        self.event_type = type(self).__name__ # allow event_type to be overridden by kw
        super().__init__(**kw)

class ToplevelResponse(MessageFromBackend):
    pass

class DebuggerResponse(MessageFromBackend):
    pass

class BackendEvent(MessageFromBackend):
    def __init__(self, event_type: str, **kw) -> None:
        super().__init__(**kw)
        self.event_type = event_type

class InlineResponse(MessageFromBackend):
    def __init__(self, command_name: str, **kw) -> None:
        super().__init__(**kw)
        self.command_name = command_name
        self.event_type = self.command_name + "_response"
        
def serialize_message(msg: Record) -> str:
    # I want to transfer only ASCII chars because encodings are not reliable 
    # (eg. can't find a way to specify PYTHONIOENCODING for cx_freeze'd program) 
    return MESSAGE_MARKER + repr(msg).encode("UTF-7").decode("ASCII") 

def parse_message(msg_string: str) -> Record:
    # DataFrames may have nan 
    # pylint: disable=unused-variable
    nan = float("nan")  # @UnusedVariable
    assert msg_string[0] == MESSAGE_MARKER
    return eval(msg_string[1:].encode("ASCII").decode("UTF-7"))

def actual_path(name: str) -> str:
    """In Windows return the path with the case it is stored in the filesystem"""
    assert os.path.isabs(name)
    assert os.path.exists(name)
    
    if os.name == "nt":
        name = os.path.realpath(name)
        
        from ctypes import create_unicode_buffer, windll

        buf = create_unicode_buffer(512)
        windll.kernel32.GetShortPathNameW(name, buf, 512)  # @UndefinedVariable
        windll.kernel32.GetLongPathNameW(buf.value, buf, 512)  # @UndefinedVariable
        assert len(buf.value) >= 2
        
        result = buf.value
        assert isinstance(result, str) 
        
        if result[1] == ":":
            # ensure drive letter is capital
            return result[0].upper() + result[1:]
        else:
            return result
    else:
        return os.path.realpath(name)

def is_same_path(name1: str, name2: str) -> bool:
    return os.path.realpath(os.path.normcase(name1)) == os.path.realpath(os.path.normcase(name2))

def path_startswith(child_name: str, dir_name: str) -> bool:
    normchild = os.path.realpath(os.path.normcase(child_name))
    normdir = os.path.realpath(os.path.normcase(dir_name))
    return normdir == normchild or normchild.startswith(normdir.rstrip(os.path.sep) + os.path.sep)

def read_source(filename):
    with tokenize.open(filename) as fp:
        return fp.read()
    

class UserError(RuntimeError):
    """Errors of this class are meant to be presented without stacktrace"""
    pass

if __name__ == "__main__":
    print(repr(actual_path("c:\\users/aivar/DesKTOp")))



from thonny import get_workbench
from thonny.workbench import SyntaxThemeSettings


def tomorrow() -> SyntaxThemeSettings:
    # https://github.com/chriskempson/tomorrow-theme/blob/master/GEdit/Tomorrow.xml
    normal_fg = "#4D4D4C"
    string_fg = "#718C00"
    
    return {
        "TEXT"          : {"foreground" : normal_fg, "insertbackground" : normal_fg, 
                           "background" : "#FFFFFF"},
        "GUTTER"        : {"foreground" : "#4d4d4c", "background" : "#efefef"},
        "current_line"  : {"background" : "#efefef"},
        "sel"           : {"foreground" : "#4D4D4C", "background" : "#D6D6D6"},
        
        "number"        : {"foreground" : "#718c00"},
        "definition"    : {"foreground" : "#4271AE", "font" : "BoldEditorFont"},
        "string"        : {"foreground" : string_fg},
        "string3"       : {"foreground" : string_fg},
        "open_string"   : {"foreground" : string_fg},
        "open_string3"  : {"foreground" : string_fg},
        "keyword"       : {"foreground" : "#8959A8", "font" : "BoldEditorFont"},
        "builtin"       : {"foreground" : "#4271ae"},
        "comment"       : {"foreground" : "#8E908C"},
    }

def tomorrow_night() -> SyntaxThemeSettings:
    # https://github.com/chriskempson/tomorrow-theme/blob/master/GEdit/Tomorrow-Night.xml
    normal_fg = "#c5c8c6"
    string_fg = "#b5bd68"
    string3_fg = string_fg
    
    return {
        "TEXT"          : {"foreground" : normal_fg, "insertbackground" : normal_fg, 
                           "background" : "#1d1f21"},
        "GUTTER"        : {"foreground" : "#c5c8c6", "background" : "#282a2e"},
        "sel"           : {"foreground" : "#c5c8c6", "background" : "#373b41"},
        "current_line"  : {"background" : "#282a2e"},
        
        "number"        : {"foreground" : "#de935f"},
        "definition"    : {"foreground" : "#81a2be", "font" : "BoldEditorFont"},
        "string"        : {"foreground" : string_fg},
        "string3"       : {"foreground" : string_fg},
        "open_string"   : {"foreground" : string3_fg},
        "open_string3"  : {"foreground" : string3_fg},
        "keyword"       : {"foreground" : "#b294bb", "font" : "BoldEditorFont"},
        "builtin"       : {"foreground" : "#81a2be"},
        "comment"       : {"foreground" : "#969896"},
        
        # paren matcher
        "surrounding_parens"  : {"background" : "#373b41", "foreground" : "white",
                                 "font" : "BoldEditorFont"},
    }

def tomorrow_night_blue() -> SyntaxThemeSettings:
    # https://github.com/chriskempson/tomorrow-theme/blob/master/GEdit/Tomorrow-Night-Blue.xml
    normal_fg = "#ffffff"
    string_fg = "#d1f1a9"
    string3_fg = string_fg
    
    return {
        "TEXT"          : {"foreground" : normal_fg, "insertbackground" : normal_fg, 
                           "background" : "#002451"},
        "GUTTER"        : {"foreground" : "#ffffff", "background" : "#00346e"},
        "sel"           : {"foreground" : "#ffffff", "background" : "#003f8e"},
        "current_line"  : {"background" : "#00346e"},
        
        "number"        : {"foreground" : "#ffc58f"},
        "definition"    : {"foreground" : "#bbdaff", "font" : "BoldEditorFont"},
        "string"        : {"foreground" : string_fg},
        "string3"       : {"foreground" : string_fg},
        "open_string"   : {"foreground" : string3_fg},
        "open_string3"  : {"foreground" : string3_fg},
        "keyword"       : {"foreground" : "#ebbbff", "font" : "BoldEditorFont"},
        "builtin"       : {"foreground" : "#bbdaff"},
        "comment"       : {"foreground" : "#7285b7"},
        
        # paren matcher
        "surrounding_parens"  : {"background" : "#00346e", "foreground" : "white",
                                 "font" : "BoldEditorFont"},
    }

def tomorrow_night_bright() -> SyntaxThemeSettings:
    # https://github.com/chriskempson/tomorrow-theme/blob/master/GEdit/Tomorrow-Night-Bright.xml
    normal_fg = "#dedede"
    string_fg = "#b9ca4a"
    string3_fg = string_fg
    
    return {
        "TEXT"          : {"foreground" : normal_fg, "insertbackground" : normal_fg, 
                           "background" : "#000000"},
        "GUTTER"        : {"foreground" : "#dedede", "background" : "#2a2a2a"},
        "current_line"  : {"background" : "#2a2a2a"},
        "sel"           : {"foreground" : "#dedede", "background" : "#424242"},
        
        "number"        : {"foreground" : "#e78c45"},
        "definition"    : {"foreground" : "#7aa6da", "font" : "BoldEditorFont"},
        "string"        : {"foreground" : string_fg},
        "string3"       : {"foreground" : string_fg},
        "open_string"   : {"foreground" : string3_fg},
        "open_string3"  : {"foreground" : string3_fg},
        "keyword"       : {"foreground" : "#c397d8", "font" : "BoldEditorFont"},
        "builtin"       : {"foreground" : "#7aa6da"},
        "comment"       : {"foreground" : "#969896"},
        
        # paren matcher
        "surrounding_parens"  : {"background" : "#2a2a2a", "foreground" : "white",
                                 "font" : "BoldEditorFont"},
    }

def tomorrow_night_eighties() -> SyntaxThemeSettings:
    # https://github.com/chriskempson/tomorrow-theme/blob/master/GEdit/Tomorrow-Night-Eighties.xml
    normal_fg = "#cccccc"
    string_fg = "#99cc99"
    string3_fg = string_fg
    
    return {
        "TEXT"          : {"foreground" : normal_fg, "insertbackground" : normal_fg, 
                           "background" : "#2d2d2d"},
        "GUTTER"        : {"foreground" : "#cccccc", "background" : "#393939"},
        "current_line"  : {"background" : "#393939"},
        "sel"           : {"foreground" : "#cccccc", "background" : "#515151"},
        
        "number"        : {"foreground" : "#f99157"},
        "definition"    : {"foreground" : "#6699cc", "font" : "BoldEditorFont"},
        "string"        : {"foreground" : string_fg},
        "string3"       : {"foreground" : string_fg},
        "open_string"   : {"foreground" : string3_fg},
        "open_string3"  : {"foreground" : string3_fg},
        "keyword"       : {"foreground" : "#cc99cc", "font" : "BoldEditorFont"},
        "builtin"       : {"foreground" : "#6699cc"},
        "comment"       : {"foreground" : "#999999"},
        
        # paren matcher
        "surrounding_parens"  : {"background" : "#393939", "foreground" : "white",
                                 "font" : "BoldEditorFont"},
    }

def load_plugin() -> None:
    get_workbench().add_syntax_theme("Tomorrow", "Default Light", tomorrow)
    get_workbench().add_syntax_theme("Tomorrow Night", "Default Dark", tomorrow_night)
    get_workbench().add_syntax_theme("Tomorrow Night Blue", "Tomorrow Night", tomorrow_night_blue)
    get_workbench().add_syntax_theme("Tomorrow Night Bright", "Tomorrow Night", tomorrow_night_bright)
    get_workbench().add_syntax_theme("Tomorrow Night Eighties", "Tomorrow Night", tomorrow_night_eighties())
    
    get_workbench().set_default("view.syntax_theme", "Default Light")



import os.path
import tkinter as tk
from tkinter import ttk

from thonny import get_workbench, misc_utils
from thonny.ui_utils import TreeFrame

_dummy_node_text = "..."
    


class BaseFileBrowser(TreeFrame):
    def __init__(self, master, show_hidden_files=False, last_folder_setting_name=None): 
        TreeFrame.__init__(self, master, 
                           ["#0", "kind", "path"], 
                           displaycolumns=(0,))
        #print(self.get_toplevel_items())
        
        self.show_hidden_files = show_hidden_files
        self.tree['show'] = ('tree',)
        
        self.hor_scrollbar = ttk.Scrollbar(self, orient=tk.HORIZONTAL)
        self.tree.config(xscrollcommand=self.hor_scrollbar.set)
        self.hor_scrollbar['command'] = self.tree.xview
        self.hor_scrollbar.grid(row=1, column=0, sticky="nsew")
        
        wb = get_workbench()
        self.folder_icon = wb.get_image("folder")
        self.python_file_icon = wb.get_image("python-file")
        self.text_file_icon = wb.get_image("text-file")
        self.generic_file_icon = wb.get_image("generic-file")
        self.hard_drive_icon = wb.get_image("hard-drive")
        
        self.tree.column('#0', width=500, anchor=tk.W)
        
        # set-up root node
        self.tree.set("", "kind", "root")
        self.tree.set("", "path", "")
        self.refresh_tree()
        
        
        self.tree.bind("<<TreeviewOpen>>", self.on_open_node)
        
        self._last_folder_setting_name = last_folder_setting_name
        self.open_initial_folder()
    
    def open_initial_folder(self):
        if self._last_folder_setting_name:
            path = get_workbench().get_option(self._last_folder_setting_name)
            if path:
                self.open_path_in_browser(path, True)
    
    def save_current_folder(self):
        if not self._last_folder_setting_name:
            return
        
        path = self.get_selected_path()
        
        if not path:
            return
        
        if os.path.isfile(path):
            path = os.path.dirname(path)
        get_workbench().set_option(self._last_folder_setting_name, path)
    
    def on_open_node(self, event):
        node_id = self.get_selected_node()
        if node_id:
            self.refresh_tree(node_id, True)
    

    
    def get_selected_node(self):
        nodes = self.tree.selection()
        assert len(nodes) <= 1
        if len(nodes) == 1:
            return nodes[0]
        else:
            return None
    
    def get_selected_path(self):
        node_id = self.get_selected_node()
        
        if node_id:
            return self.tree.set(node_id, "path")
        else:
            return None
    
    def open_path_in_browser(self, path, see=True):
        
        # unfortunately os.path.split splits from the wrong end (for this case)
        def split(path):
            head, tail = os.path.split(path)
            if head == "" and tail == "":
                return []
            elif head == path or tail == path:
                return [path]
            elif head == "":
                return [tail]
            elif tail == "":
                return split(head)
            else:
                return split(head) + [tail]
        
        parts = split(path)
        current_node_id = ""
        current_path = ""
        
        while parts != []:
            current_path = os.path.join(current_path, parts.pop(0))
            
            for child_id in self.tree.get_children(current_node_id):
                child_path = self.tree.set(child_id, "path")
                if child_path == current_path:
                    self.tree.item(child_id, open=True)
                    self.refresh_tree(child_id)
                    current_node_id = child_id
                    break
        
        if see:
            self.tree.selection_set(current_node_id)
            self.tree.focus(current_node_id)
            
            if self.tree.set(current_node_id, "kind") == "file":
                self.tree.see(self.tree.parent(current_node_id))
            else:
                self.tree.see(current_node_id)    
                
    
    def refresh_tree(self, node_id="", opening=None):
        path = self.tree.set(node_id, "path")
        #print("REFRESH", path)
        
        if os.path.isfile(path):
            self.tree.set_children(node_id)
            self.tree.item(node_id, open=False)
        else:
            # either root or directory
            if node_id == "" or self.tree.item(node_id, "open") or opening == True:
                fs_children_names = self.listdir(path, self.show_hidden_files)
                tree_children_ids = self.tree.get_children(node_id) 
                
                # recollect children
                children = {}
                
                # first the ones, which are present already in tree
                for child_id in tree_children_ids:
                    name = self.tree.item(child_id, "text")
                    if name in fs_children_names:
                        children[name] = child_id
                
                # add missing children
                for name in fs_children_names:
                    if name not in children:
                        children[name] = self.tree.insert(node_id, "end")
                        self.tree.set(children[name], "path", os.path.join(path, name))
                        
                
                def file_order(name):
                    # items in a folder should be ordered so that 
                    # folders come first and names are ordered case insensitively
                    return (os.path.isfile(os.path.join(path, name)),
                            name.upper())
                
                # update tree
                ids_sorted_by_name = list(map(lambda key: children[key],
                                              sorted(children.keys(), key=file_order)))
                self.tree.set_children(node_id, *ids_sorted_by_name)
                
                for child_id in ids_sorted_by_name:
                    self.update_node_format(child_id)
                    self.refresh_tree(child_id)
                
            else:
                # closed dir
                # Don't fetch children yet, but ensure that expand button is visible
                children_ids = self.tree.get_children(node_id)
                if len(children_ids) == 0:
                    self.tree.insert(node_id, "end", text=_dummy_node_text) 
    
    def update_node_format(self, node_id):
        assert node_id != ""
        
        path = self.tree.set(node_id, "path")
        
        
        if os.path.isdir(path) or path.endswith(":") or path.endswith(":\\"):
            self.tree.set(node_id, "kind", "dir")
            if path.endswith(":") or path.endswith(":\\"):
                img = self.hard_drive_icon
            else:
                img = self.folder_icon
        else:
            self.tree.set(node_id, "kind", "file")
            if path.lower().endswith(".py"):
                img = self.python_file_icon
            elif path.lower().endswith(".txt") or path.lower().endswith(".csv"):
                img = self.text_file_icon
            else:
                img = self.generic_file_icon
            
        # compute caption 
        text = os.path.basename(path)
        if text == "": # in case of drive roots 
            text = path
            
        self.tree.item(node_id, text=" " + text, image=img)
        self.tree.set(node_id, "path", path)
    
    def listdir(self, path="", include_hidden_files=False):
        
        if path == "" and misc_utils.running_on_windows():
            result = misc_utils.get_win_drives()
        else:
            if path == "":
                first_level = True
                path = "/"
            else:
                first_level = False
            result = [x for x in os.listdir(path) 
                        if include_hidden_files 
                            or not misc_utils.is_hidden_or_system_file(os.path.join(path, x))]
            
            if first_level:
                result = ["/" + x for x in result]
            
        return sorted(result, key=str.upper)



import tkinter

from thonny.plugins.paren_matcher import ParenMatcher

TEST_STR1 = """age = int(input("Enter age: "))
if age > 18:
    l = ["H", "I"]
    print(l)
else:
    print("Hello!", end='')
    print("What's your name?")
"""


def test_regular_closed():
    insert_pos_groups = (("1.9", "1.10", "1.13", "1.31"),
                         ("1.30", "1.29", "1.25", "1.15"))
    expected_indices = (("1.9", "1.30", []),
                        ("1.15", "1.29", []))

    text_widget = tkinter.Text()
    text_widget.insert("end", TEST_STR1)

    matcher = ParenMatcher(text_widget)
    matcher.text = text_widget
    for i, group in enumerate(insert_pos_groups):
        for insert_pos in group:
            text_widget.mark_set("insert", insert_pos)

            actual = matcher.find_surrounding("1.0", "end")
            expected = expected_indices[i]

            assert actual == expected, "\nExpected: %s\nGot: %s" % (expected, actual)
        print("\rPassed %d of %d" % (i+1, len(insert_pos_groups)), end="")




class OneWire:
    ''
    MATCH_ROM = 85
    SEARCH_ROM = 240
    SKIP_ROM = 204
    def _search_rom():
        pass

    def crc8():
        pass

    def readbit():
        pass

    def readbyte():
        pass

    def readinto():
        pass

    def reset():
        pass

    def scan():
        pass

    def select_rom():
        pass

    def write():
        pass

    def writebit():
        pass

    def writebyte():
        pass


class OneWireError:
    ''
_ow = None
def const():
    pass



import os.path
import sys
from typing import TYPE_CHECKING, cast, Optional


if os.environ.get("THONNY_USER_DIR", ""):
    THONNY_USER_DIR = os.environ["THONNY_USER_DIR"]
elif (hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix
      or hasattr(sys, 'real_prefix') and getattr(sys, "real_prefix") != sys.prefix):
    # we're in a virtualenv or venv
    THONNY_USER_DIR = os.path.join(sys.prefix, ".thonny")
else:
    THONNY_USER_DIR = os.path.expanduser(os.path.join("~", ".thonny"))
    
    
def launch():
    _prepare_thonny_user_dir()
    _misc_prepare()
    
    try:
        from thonny import workbench
        
        if _should_delegate():
            # First check if there is existing Thonny instance to handle the request
            delegation_result = _try_delegate_to_existing_instance(sys.argv[1:])
            if delegation_result == True: #pylint: disable=singleton-comparison
                # we're done
                print("Delegated to an existing Thonny instance. Exiting now.")
                return 0
            
            if hasattr(delegation_result, "accept"):
                # we have server socket to put in use
                server_socket = delegation_result
            else:
                server_socket = None
                 
            bench = workbench.Workbench(server_socket)
        else:
            bench = workbench.Workbench()
            
        try:
            bench.mainloop()
        except SystemExit:
            bench.destroy()
        return 0
    except SystemExit as e:
        from tkinter import messagebox
        messagebox.showerror("System exit", str(e))
        return -1
    except Exception:
        from logging import exception
        exception("Internal launch or mainloop error")
        import tkinter.messagebox
        import traceback
        tkinter.messagebox.showerror("Internal error", traceback.format_exc())
        return -1
    finally:
        runner = get_runner()
        if runner is not None:
            runner.destroy_backend()
    
    return 0

def _prepare_thonny_user_dir():
    if not os.path.exists(THONNY_USER_DIR):
        os.makedirs(THONNY_USER_DIR, mode=0o700, exist_ok=True)
        
        # user_dir_template is a post-installation means for providing
        # alternative default user environment in multi-user setups
        template_dir = os.path.join(os.path.dirname(__file__), "user_dir_template")
                    
        if os.path.isdir(template_dir):
            import shutil
            
            def copy_contents(src_dir, dest_dir):
                # I want the copy to have current user permissions
                for name in os.listdir(src_dir):
                    src_item = os.path.join(src_dir, name)
                    dest_item = os.path.join(dest_dir, name)
                    if os.path.isdir(src_item):
                        os.makedirs(dest_item, mode=0o700)
                        copy_contents(src_item, dest_item)
                    else:
                        shutil.copyfile(src_item, dest_item)
                        os.chmod(dest_item, 0o600)
                        
            copy_contents(template_dir, THONNY_USER_DIR)
            

def _should_delegate():
    from thonny import workbench
    from thonny.config import try_load_configuration
    configuration_manager = try_load_configuration(workbench.CONFIGURATION_FILE_NAME)
    # Setting the default
    configuration_manager.set_default("general.single_instance", workbench.SINGLE_INSTANCE_DEFAULT)
    # getting the value (may use the default or return saved value)
    return configuration_manager.get_option("general.single_instance")

def _try_delegate_to_existing_instance(args):
    import socket
    from thonny import workbench
    try:
        # Try to create server socket.
        # This is fastest way to find out if Thonny is already running
        serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        serversocket.bind(("localhost", workbench.THONNY_PORT))
        serversocket.listen(10)
        # we were able to create server socket (ie. Thonny was not running)
        # Let's use the socket in Thonny so that requests coming while 
        # UI gets constructed don't get lost.
        # (Opening several files with Thonny in Windows results in many
        # Thonny processes opened quickly) 
        return serversocket
    except OSError:
        # port was already taken, most likely by previous Thonny instance.
        # Try to connect and send arguments
        try:
            return _delegate_to_existing_instance(args)
        except Exception:
            import traceback
            traceback.print_exc()
            return False
        
        
def _delegate_to_existing_instance(args):
    import socket
    from thonny import workbench
    data = repr(args).encode(encoding='utf_8')
    sock = socket.create_connection(("localhost", workbench.THONNY_PORT))
    sock.sendall(data)
    sock.shutdown(socket.SHUT_WR)
    response = bytes([])
    while len(response) < len(workbench.SERVER_SUCCESS):
        new_data = sock.recv(2)
        if len(new_data) == 0:
            break
        else:
            response += new_data
    
    return response.decode("UTF-8") == workbench.SERVER_SUCCESS

def _misc_prepare():
    if os.name == "nt":
        import ctypes
        ctypes.windll.user32.SetProcessDPIAware()    
    
def get_version():
    try:
        package_dir = os.path.dirname(sys.modules["thonny"].__file__)
        with open(os.path.join(package_dir, "VERSION"), encoding="ASCII") as fp:
            return fp.read().strip()
    except Exception:
        return "0.0.0"

if TYPE_CHECKING:
    # Following imports are required for MyPy
    # http://mypy.readthedocs.io/en/stable/common_issues.html#import-cycles
    import thonny.workbench
    from thonny.workbench import Workbench
    from thonny.running import Runner
    from thonny.shell import ShellView
          
_workbench = None # type: Optional[Workbench]
def get_workbench() -> 'Workbench':
    return cast('Workbench', _workbench)

_runner = None # type: Optional[Runner]
def get_runner() -> 'Runner':    
    return cast('Runner', _runner)

def get_shell() -> 'ShellView':
    return cast('ShellView', get_workbench().get_view("ShellView"))



import tkinter as tk
from tkinter import ttk

from thonny import get_runner, get_workbench, misc_utils, ui_utils
from thonny.codeview import CodeView


class ShellMacroDialog(tk.Toplevel):
    def __init__(self, master):
        tk.Toplevel.__init__(self, master)
        

        self.title("Configure shell macro")
        if misc_utils.running_on_mac_os():
            self.configure(background="systemSheetBackground")
        self.transient(master)
        self.grab_set() # to make it active
        #self.grab_release() # to allow eg. copy something from the editor 
        
        self._create_widgets()
        
        self.bind('<Escape>', self._on_close, True) 
        self.protocol("WM_DELETE_WINDOW", self._on_close)
        ui_utils.center_window(self, master)
        self.main_command_text.focus_set()
        
    def _create_widgets(self):
        
        bg = "#ffff99"
        banner_frame = tk.Frame(self, background=bg)
        banner_frame.grid(row=0, column=0, sticky="nsew")
        banner_frame.rowconfigure(0, weight=1)
        banner_frame.columnconfigure(0, weight=1)
        banner_text = tk.Label(banner_frame, text="These\nare\ninstructions asdfa afs fa sfasdf",
                                background=bg, justify="left")
        banner_text.grid(column=0, row=0, pady=10, padx=10, sticky="nsew")
        
        
        main_frame = ttk.Frame(self)
        main_frame.grid(row=1, column=0, sticky=tk.NSEW, padx=15, pady=15)
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        
        self.main_command_text = CodeView(main_frame, height=5)
        self.main_command_text.grid(column=0, row=1, sticky="nsew")
        #main_command_text["relief"] = "groove"
        
        main_frame.rowconfigure(1, weight=1)
        main_frame.columnconfigure(0, weight=1)
        
        button_frame = ttk.Frame(main_frame)
        button_frame.grid(row=2, column=0, sticky="nsew")
        
        run_button = ttk.Button(button_frame, text="Save and execute", command=self._save_exec)
        run_button.grid(row=0, column=1, sticky="nsew")
        ok_button = ttk.Button(button_frame, text="Save", command=self._save)
        ok_button.grid(row=0, column=2, sticky="nsew")
        cancel_button = ttk.Button(button_frame, text="Cancel", command=self._cancel)
        cancel_button.grid(row=0, column=3, sticky="nsew")
        button_frame.columnconfigure(0, weight=1)

    def _on_close(self, event=None):
        self.destroy()
        
    def _save_exec(self, event=None):
        self._save(event)
        execute_macro()
        
    def _save(self, event=None):
        source = self.main_command_text.text.get("1.0", "end")
        get_workbench().set_option("run.shell_macro_main", repr(source))
        self.destroy()
    
    def _cancel(self, event=None):
        self.destroy()
        
        

def show_dialog():
    dlg = ShellMacroDialog(get_workbench())
    ui_utils.show_dialog(dlg)

def execute_macro():
    if get_runner().is_waiting_toplevel_command():
        source = get_workbench().get_option("run.shell_macro_main")
        if source is not None:
            shell = get_workbench().show_view("ShellView")
            shell.submit_python_code(source.strip() + "\n")

def _load_plugin():
    get_workbench().set_default("run.shell_macro_main", None)
    get_workbench().add_command("configure_shell_macro", "run", "Configure shell macro...",
                                show_dialog)
    get_workbench().add_command("execute_shell_macro", "run", "Execute shell macro",
                                execute_macro, default_sequence="<F9>")



import astroid

from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

class FileNamingChecker(BaseChecker):
    __implements__ = IAstroidChecker

    name = 'file-naming'
    priority = -1
    msgs = {
        'W7401': (
            'Script shadows a library module.',
            'non-unique-returns',
            'All constants returned in a function should be unique.'
        ),
    }
    options = (
        (
            'ignore-ints',
            {
                'default': False, 'type': 'yn', 'metavar' : '<y_or_n>',
                'help': 'Allow returning non-unique integers',
            }
        ),
    )


def acos():
    pass

def acosh():
    pass

def asin():
    pass

def asinh():
    pass

def atan():
    pass

def atan2():
    pass

def atanh():
    pass

def ceil():
    pass

def copysign():
    pass

def cos():
    pass

def cosh():
    pass

def degrees():
    pass

e = 2.718282
def erf():
    pass

def erfc():
    pass

def exp():
    pass

def expm1():
    pass

def fabs():
    pass

def floor():
    pass

def fmod():
    pass

def frexp():
    pass

def gamma():
    pass

def isfinite():
    pass

def isinf():
    pass

def isnan():
    pass

def ldexp():
    pass

def lgamma():
    pass

def log():
    pass

def log10():
    pass

def log2():
    pass

def modf():
    pass

pi = 3.141593
def pow():
    pass

def radians():
    pass

def sin():
    pass

def sinh():
    pass

def sqrt():
    pass

def tan():
    pass

def tanh():
    pass

def trunc():
    pass
